/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.CCompiler;
import info.yeppp.ebuilda.generic.CppCompiler;

import java.util.*;

public abstract class AbstractCppCompiler implements CppCompiler {

	protected AbstractCppCompiler(ABI abi) {
		this.abi = abi;
	}

	protected AbstractCppCompiler(AbstractCppCompiler prototype) {
		this.abi = prototype.abi;
		this.sourceDirectory = prototype.sourceDirectory;
		this.objectDirectory = prototype.objectDirectory;
		this.macros = new HashMap<String, String>(prototype.macros);
		this.includeDirectories = new HashSet<AbsoluteDirectoryPath>(prototype.includeDirectories);
		this.globalIncludeDirectories = new HashSet<AbsoluteDirectoryPath>(prototype.globalIncludeDirectories);
		this.sysRoot = prototype.sysRoot;
		this.optimization = prototype.optimization;
		this.globalOptimization = prototype.globalOptimization;
		this.debugInformation = prototype.debugInformation;
		this.debugEnabled = prototype.debugEnabled;
		this.verboseBuild = prototype.verboseBuild;
		this.pic = prototype.pic;
		this.runtimeLibrary = prototype.runtimeLibrary;
	}

	protected final String getExtension() {
		if (getABI().getOperatingSystem().equals(OperatingSystem.Windows)) {
			return "obj";
		} else {
			return "o";
		}
	}

	@Override
	public AbsoluteFilePath getObjectPath(AbsoluteFilePath sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (objectDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final RelativeFilePath relativeSourcePath = sourcePath.getRelativePath(sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		return new AbsoluteFilePath(objectDirectory,  relativeSourcePath.replaceExtension(getExtension()));
	}

	@Override
	public void setSourceDirectory(AbsoluteDirectoryPath sourceDirectory) {
		this.sourceDirectory = sourceDirectory;
	}

	@Override
	public AbsoluteDirectoryPath getSourceDirectory() {
		return this.sourceDirectory;
	}

	@Override
	public void setObjectDirectory(AbsoluteDirectoryPath objectDirectory) {
		this.objectDirectory = objectDirectory;
	}

	@Override
	public AbsoluteDirectoryPath getObjectDirectory() {
		return this.objectDirectory;
	}

	@Override
	public final void setSysRoot(AbsoluteDirectoryPath sysRoot) {
		this.sysRoot = sysRoot;
	}

	@Override
	public final AbsoluteDirectoryPath getSysRoot() {
		return this.sysRoot;
	}

	@Override
	public void setVerboseBuild(boolean verboseBuild) {
		this.verboseBuild = verboseBuild;
	}

	@Override
	public boolean isVerboseBuild() {
		return this.verboseBuild;
	}

	@Override
	public void setOptimization(CCompiler.Optimization optimization) {
		this.optimization = optimization;
	}

	@Override
	public CCompiler.Optimization getOptimization() {
		return this.optimization;
	}

	@Override
	public void setGlobalOptimization(boolean globalOptimization) {
		this.globalOptimization = globalOptimization;
	}

	@Override
	public boolean getGlobalOptimization() {
		return this.globalOptimization;
	}

	@Override
	public void setDebugInformation(boolean debugInformation) {
		this.debugInformation = debugInformation;
	}

	@Override
	public boolean getDebugInformation() {
		return this.debugInformation;
	}

	@Override
	public void setDebugBuild(boolean debugEnabled) {
		this.debugEnabled = debugEnabled;
	}

	@Override
	public boolean isDebugBuild() {
		return this.debugEnabled;
	}

	@Override
	public void setRuntimeLibrary(CCompiler.RuntimeLibrary runtimeLibrary) {
		this.runtimeLibrary = runtimeLibrary;
	}

	@Override
	public CCompiler.RuntimeLibrary getRuntimeLibrary() {
		return this.runtimeLibrary;
	}

	@Override
	public void setPositionIndependentCodeGeneration(PositionIndependentCodeGeneration pic) {
		this.pic = pic;
	}

	@Override
	public PositionIndependentCodeGeneration getPositionIndependentCodeGeneration() {
		return this.pic;
	}

	@Override
	public boolean addMacro(String name) {
		return this.addMacro(name, "1");
	}

	@Override
	public boolean addMacro(String name, String value) {
		if (macros.containsKey(name)) {
			return false;
		} else {
			macros.put(name, value);
			return true;
		}
	}

	@Override
	public boolean removeMacro(String name) {
		return (macros.remove(name) != null);
	}

	@Override
	public void clearMacros() {
		macros.clear();
	}

	@Override
	public boolean addGlobalIncludeDirectory(AbsoluteDirectoryPath includeDirectory) {
		return globalIncludeDirectories.add(includeDirectory);
	}

	@Override
	public boolean addGlobalIncludeDirectories(Collection<AbsoluteDirectoryPath> includeDirectories) {
		return globalIncludeDirectories.addAll(includeDirectories);
	}

	@Override
	public boolean removeGlobalIncludeDirectory(AbsoluteDirectoryPath includeDirectory) {
		return globalIncludeDirectories.remove(includeDirectory);
	}

	@Override
	public void clearGlobalIncludeDirectories() {
		globalIncludeDirectories.clear();
	}

	@Override
	public boolean addIncludeDirectory(AbsoluteDirectoryPath includePath) {
		return includeDirectories.add(includePath);
	}

	@Override
	public boolean removeIncludeDirectory(AbsoluteDirectoryPath includePath) {
		return includeDirectories.remove(includePath);
	}

	@Override
	public void clearIncludeDirectories() {
		includeDirectories.clear();
	}

	@Override
	public void clearAll() {
		this.clearGlobalIncludeDirectories();
		this.clearIncludeDirectories();
		this.clearMacros();
	}

	@Override
	public Boolean isRttiEnabled() {
		return this.rttiEnabled;
	}

	@Override
	public void setRttiEnabled(Boolean rttiEnabled) {
		this.rttiEnabled = rttiEnabled;
	}

	@Override
	public void setExceptionsSupport(Exceptions exceptions) {
		this.exceptions = exceptions;
	}

	@Override
	public Exceptions getExceptionsEnabled() {
		return this.exceptions;
	}

	@Override
	public ABI getABI() {
		return this.abi;
	}

	private final ABI abi;
	protected AbsoluteDirectoryPath sourceDirectory = null;
	protected AbsoluteDirectoryPath objectDirectory = null;
	protected Map<String, String> macros = new HashMap<String, String>();
	protected Set<AbsoluteDirectoryPath> includeDirectories = new HashSet<AbsoluteDirectoryPath>();
	protected Set<AbsoluteDirectoryPath> globalIncludeDirectories = new HashSet<AbsoluteDirectoryPath>();
	protected AbsoluteDirectoryPath sysRoot = null;
	protected CCompiler.Optimization optimization = CCompiler.Optimization.NoOptimization;
	protected boolean globalOptimization;
	protected boolean debugInformation = true;
	protected boolean debugEnabled = false;
	protected boolean verboseBuild = false;
	protected PositionIndependentCodeGeneration pic = PositionIndependentCodeGeneration.NoPIC;
	protected CCompiler.RuntimeLibrary runtimeLibrary = CCompiler.RuntimeLibrary.DynamicRuntimeLibrary;
	protected Exceptions exceptions = Exceptions.SyncronousExceptions;
	protected Boolean rttiEnabled = true;

}
