/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.JavaCompiler;

import java.util.HashSet;
import java.util.Set;

public abstract class AbstractJavaCompiler implements JavaCompiler {

	protected AbstractJavaCompiler() {
	}

	protected AbstractJavaCompiler(AbstractJavaCompiler prototype) {
		this.sourceDirectory = prototype.sourceDirectory;
		this.classDirectory = prototype.classDirectory;
		this.classPathDirectories = new HashSet<AbsoluteDirectoryPath>(prototype.classPathDirectories);
		this.debugBuild = prototype.debugBuild;
	}

	@Override
	public AbsoluteFilePath getObjectPath(AbsoluteFilePath sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (classDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final RelativeFilePath relativeSourcePath = sourcePath.getRelativePath(sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		return new AbsoluteFilePath(classDirectory,  relativeSourcePath.replaceExtension("class"));
	}

	public void setSourceDirectory(AbsoluteDirectoryPath sourceDirectory) {
		this.sourceDirectory = sourceDirectory;
	}

	public AbsoluteDirectoryPath getSourceDirectory() {
		return this.sourceDirectory;
	}

	public void setClassDirectory(AbsoluteDirectoryPath classDirectory) {
		this.classDirectory = classDirectory;
	}

	public AbsoluteDirectoryPath getClassDirectory() {
		return this.classDirectory;
	}

	public void setDebugBuild(boolean debugEnabled) {
		this.debugBuild = debugEnabled;
	}

	public boolean isDebugBuild() {
		return this.debugBuild;
	}

	public void setVerboseBuild(boolean verboseBuild) {
		this.verboseBuild = verboseBuild;
	}

	public boolean isVerboseBuild() {
		return this.verboseBuild;
	}

	public boolean addClassPathDirectory(AbsoluteDirectoryPath classPath) {
		return classPathDirectories.add(classPath);
	}

	public boolean removeClassPathDirectory(AbsoluteDirectoryPath classPath) {
		return classPathDirectories.remove(classPath);
	}

	public void clearClassPathDirectories() {
		classPathDirectories.clear();
	}

	protected AbsoluteDirectoryPath sourceDirectory = null;
	protected AbsoluteDirectoryPath classDirectory = null;
	protected Set<AbsoluteDirectoryPath> classPathDirectories = new HashSet<AbsoluteDirectoryPath>();
	protected boolean debugBuild = false;
	protected boolean verboseBuild = false;
}
