/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.generic.Linker;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public abstract class AbstractLinker implements Linker {

	public static enum Target {
		Executable,
		DynamicLibrary
	}

	public AbstractLinker(ABI abi) {
		this.abi = abi;
	}

	public AbstractLinker(AbstractLinker linker) {
		this.abi = linker.abi;
		this.debugInformation = linker.debugInformation;
		this.verboseBuild = linker.verboseBuild;
		this.runtimeLibraryUse = linker.runtimeLibraryUse;
		this.globalOptimization = linker.globalOptimization;
		this.binariesDirectory = linker.binariesDirectory;
		this.objectDirectory = linker.objectDirectory;
		this.staticLibraries = new ArrayList<String>(linker.staticLibraries);
		this.dynamicLibraries = new ArrayList<String>(linker.dynamicLibraries);
		this.globalLibraryDirectories = new ArrayList<AbsoluteDirectoryPath>(linker.globalLibraryDirectories);
		this.libraryDirectories = new ArrayList<AbsoluteDirectoryPath>(linker.libraryDirectories);
	}

	@Override
	public void setDebugInformation(boolean debugInformation) {
		this.debugInformation = debugInformation;
	}

	@Override
	public boolean getDebugInformation() {
		return this.debugInformation;
	}

	@Override
	public void setBinariesDirectory(AbsoluteDirectoryPath binariesDirectory) {
		this.binariesDirectory = binariesDirectory;
	}

	@Override
	public AbsoluteDirectoryPath getBinariesDirectory() {
		return this.binariesDirectory;
	}

	@Override
	public void setObjectDirectory(AbsoluteDirectoryPath objectDirectory) {
		this.objectDirectory = objectDirectory;
	}

	@Override
	public AbsoluteDirectoryPath getObjectDirectory() {
		return this.objectDirectory;
	}

	@Override
	public void setVerboseBuild(boolean verboseBuild) {
		this.verboseBuild = verboseBuild;
	}

	@Override
	public boolean isVerboseBuild() {
		return this.verboseBuild;
	}

	@Override
	public void setGlobalOptimization(boolean globalOptimization) {
		this.globalOptimization = globalOptimization;
	}

	@Override
	public boolean getGlobalOptimization() {
		return this.globalOptimization;
	}

	@Override
	public boolean addStaticLibraryDependence(String libraryName) {
		return staticLibraries.add(libraryName);
	}

	@Override
	public boolean removeStaticLibraryDependence(String libraryName) {
		return staticLibraries.remove(libraryName);
	}

	@Override
	public void clearStaticLibraryDependences() {
		staticLibraries.clear();
	}

	@Override
    public boolean addDynamicLibraryDependence(String libraryName) {
        return dynamicLibraries.add(libraryName);
    }

	@Override
    public boolean removeDynamicLibraryDependence(String libraryName) {
        return dynamicLibraries.remove(libraryName);
    }

	@Override
    public void clearDynamicLibraryDependences() {
        dynamicLibraries.clear();
    }

	@Override
	public PositionIndependentCodeGeneration getPIC() {
		return this.pic;
	}

	@Override
	public void setPIC(PositionIndependentCodeGeneration pic) {
		this.pic = pic;
	}

	@Override
	public void setRuntimeLibraryUse(boolean runtimeLibraryUse) {
		this.runtimeLibraryUse = runtimeLibraryUse;
	}

	@Override
	public boolean getRuntimeLibraryUse() {
		return this.runtimeLibraryUse;
	}

	@Override
	public boolean addGlobalLibraryDirectory(AbsoluteDirectoryPath libraryDirectory) {
		return globalLibraryDirectories.add(libraryDirectory);
	}

	@Override
	public void addGlobalLibraryDirectories(Collection<AbsoluteDirectoryPath> libraryDirectories) {
		globalLibraryDirectories.addAll(libraryDirectories);
	}

	@Override
	public boolean removeGlobalLibraryDirectory(AbsoluteDirectoryPath libraryDirectory) {
		return globalLibraryDirectories.remove(libraryDirectory);
	}

	@Override
	public void clearGlobalLibraryDirectories() {
		globalLibraryDirectories.clear();
	}

	@Override
	public boolean addLibraryDirectory(AbsoluteDirectoryPath libraryDirectory) {
		return libraryDirectories.add(libraryDirectory);
	}

	@Override
	public void addLibraryDirectories(Collection<AbsoluteDirectoryPath> libraryDirectories) {
		libraryDirectories.addAll(libraryDirectories);
	}

	@Override
	public boolean removeLibraryDirectory(AbsoluteDirectoryPath libraryDirectory) {
		return libraryDirectories.remove(libraryDirectory);
	}

	@Override
	public void clearLibraryDirectories() {
		libraryDirectories.clear();
	}

	@Override
	public void clearAll() {
		this.runtimeLibraryUse = true;
		this.clearDynamicLibraryDependences();
		this.clearGlobalLibraryDirectories();
		this.clearLibraryDirectories();
		this.clearStaticLibraryDependences();
	}

	@Override
	public ABI getABI() {
		return this.abi;
	}

	protected final ABI abi;
	protected boolean debugInformation = true;
	protected boolean verboseBuild = false;
	protected boolean runtimeLibraryUse = true;
	protected PositionIndependentCodeGeneration pic = PositionIndependentCodeGeneration.NoPIC;
	protected boolean globalOptimization = false;
	protected AbsoluteDirectoryPath binariesDirectory = null;
	protected AbsoluteDirectoryPath objectDirectory = null;
	protected List<String> staticLibraries = new ArrayList<String>();
	protected List<String> dynamicLibraries = new ArrayList<String>();
	protected List<AbsoluteDirectoryPath> globalLibraryDirectories = new ArrayList<AbsoluteDirectoryPath>();
	protected List<AbsoluteDirectoryPath> libraryDirectories = new ArrayList<AbsoluteDirectoryPath>();
}
