/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.generic.Versioned;
import info.yeppp.ebuilda.internals.Program;

import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class AppleDSymUtil implements Versioned {

	public static VersionList<AppleDSymUtil> enumerate(Machine machine, ABI abi) {
		final VersionList<AppleDSymUtil> dsymutils = new VersionList<AppleDSymUtil>();
		if (machine.getOperatingSystem().equals(OperatingSystem.MacOSX) && abi.getOperatingSystem().equals(OperatingSystem.MacOSX)) {
			final ToolInfo dsymutilInfo = AppleDSymUtil.verify(new AbsoluteFilePath(machine, "/usr/bin/dsymutil"));
			if (dsymutilInfo.isValid()) {
				final AppleDSymUtil dsymutil = new AppleDSymUtil(dsymutilInfo.getPath(), abi, dsymutilInfo.getVersion());
				dsymutils.add(dsymutil);
			}
		}
		return dsymutils;
	}

	private static final Pattern infoPattern = Pattern.compile("@\\(#\\)PROGRAM:dsymutil\\b.*");

	public static ToolInfo verify(AbsoluteFilePath dsymutilPath) {
		final Program.Environment environment = new Program.Environment();
		final List<String> infoStrings = Program.getStdoutLines(dsymutilPath, new Program.Parameters("-v"), environment, 1);
		if (infoStrings.isEmpty()) {
			return new ToolInfo(dsymutilPath);
		} else {
			final Matcher infoMatcher = infoPattern.matcher(infoStrings.get(0));
			if (infoMatcher.matches()) {
				return new ToolInfo(dsymutilPath, dsymutilPath.getMachine().getArchitecture().is64bit() ? ABI.X64_MacOSX_SystemV_Default : ABI.X86_MacOSX_Pic_Default,
						new ABI[]{ABI.X64_MacOSX_SystemV_Default, ABI.X86_MacOSX_Pic_Default}, new Version(0));
			} else {
				return new ToolInfo(dsymutilPath);
			}
		}
	}

	public AppleDSymUtil(AbsoluteFilePath dsymutilPath, ABI abi, Version version) {
		this.dsymutilPath = dsymutilPath;
		this.abi = abi;
		this.version = version;
		this.verboseBuild = false;
	}

	public AppleDSymUtil(AppleDSymUtil prototype) {
		this.dsymutilPath = prototype.dsymutilPath;
		this.abi = prototype.abi;
		this.version = prototype.version;
		this.verboseBuild = prototype.verboseBuild;
	}

	public BuildMessages extractDebugInformation(AbsoluteFilePath binary, AbsoluteFilePath debugInformation) {
		final BuildMessages buildMessages = new BuildMessages();
		final Program.Parameters parameters = new Program.Parameters();
		parameters.add("--flat");
		parameters.add("--out=" + debugInformation.getCanonicalForm());
		parameters.add(binary.getCanonicalForm());
		final Program.Environment environment = new Program.Environment();
		final Program.EventHandler eventHandler = new PassThroughOutputParser(buildMessages, binary);
		if (isVerboseBuild()) {
			System.out.println(dsymutilPath + " " + parameters.toString());
		}
		final Program dsymutil = new Program(dsymutilPath, parameters, environment, eventHandler);
		dsymutil.join();
		return buildMessages;
	}

	public void setVerboseBuild(boolean verboseBuild) {
		this.verboseBuild = verboseBuild;
	}

	public boolean isVerboseBuild() {
		return this.verboseBuild;
	}

	public ABI getABI() {
		return this.abi;
	}

	@Override
	public Version getVersion() {
		return this.version;
	}

	private final AbsoluteFilePath dsymutilPath;
	private final ABI abi;
	protected final Version version;
	private boolean verboseBuild;
}
