/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.generic.CppCompiler;
import info.yeppp.ebuilda.generic.Versioned;
import info.yeppp.ebuilda.internals.Program;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class ClangCppCompiler extends AbstractCppCompiler implements Versioned {

	private static final Pattern versionPattern = Pattern.compile("^(?:Ubuntu )?clang version (\\d+(?:\\.\\d+)+)\\b");

	public static ToolInfo verify(AbsoluteFilePath compilerPath) {
		Program.Parameters parameters = new Program.Parameters("-v");
		Program.Environment environment = new Program.Environment();
		final List<String> outputStrings = Program.getStderrLines(compilerPath, parameters, environment, 2);
		if (outputStrings.size() < 2) {
			return new ToolInfo(compilerPath);
		} else {
			final String versionString = outputStrings.get(0);
			final Matcher versionMatcher = versionPattern.matcher(versionString);
			final String targetString = outputStrings.get(1);
			ABI defaultABI = null;
			ABI[] supportedABI = null;
			if (versionMatcher.find() && targetString.startsWith("Target: ")) {
				final String target = targetString.substring(8);
				if (target.matches("x86_64\\-(\\w+\\-)?linux(\\-gnu)?")) {
					defaultABI = ABI.X64_Linux_SystemV_Default;
					supportedABI = new ABI[]{ ABI.X64_Linux_SystemV_Default, ABI.X86_Linux_Pic_i586 };
				} else if (target.matches("x86_64\\-apple\\-darwin(?:\\d+(?:\\.\\d+)+)")) {
					defaultABI = ABI.X64_MacOSX_SystemV_Default;
					supportedABI = new ABI[]{ ABI.X64_MacOSX_SystemV_Default, ABI.X86_MacOSX_Pic_Default };
				} else if (target.matches("arm\\-(\\w+\\-)?linux\\-gnueabihf")) {
					defaultABI = ABI.ARM_Linux_HardEABI_V7A;
					supportedABI = new ABI[]{ ABI.ARM_Linux_HardEABI_V7A };
				} else if (target.matches("arm\\-(\\w+\\-)?linux\\-gnueabi")) {
					defaultABI = ABI.ARM_Linux_SoftEABI_V5T;
					supportedABI = new ABI[]{ ABI.ARM_Linux_SoftEABI_V5T };
				}
				final Version version = new Version(versionMatcher.group(1));
				return new ToolInfo(compilerPath, defaultABI, supportedABI, version);
			} else {
				return new ToolInfo(compilerPath);
			}
		}
	}

	ClangCppCompiler(AbsoluteFilePath compilerPath, ABI abi, Version version, Collection<AbsoluteDirectoryPath> defaultIncludeDirectories, AbsoluteDirectoryPath sysRoot) {
		super(abi);
		this.compilerPath = compilerPath;
		this.defaultIncludeDirectories = new ArrayList<AbsoluteDirectoryPath>(defaultIncludeDirectories);
		this.sysRoot = sysRoot;
		this.version = version;
	}

	public ClangCppCompiler(ClangCppCompiler prototype) {
		super(prototype);
		this.compilerPath = prototype.compilerPath;
		this.defaultIncludeDirectories = new ArrayList<AbsoluteDirectoryPath>(prototype.defaultIncludeDirectories);
		this.version = prototype.version;
		this.sysRoot = prototype.sysRoot;
		this.pic = prototype.pic;
	}

	public GccCompatibleCompilerLinker asLinker(Collection<AbsoluteDirectoryPath> defaultGlobalLibraryDirectories) {
		return new GccCompatibleCompilerLinker(this.compilerPath, getABI(), getVersion(), defaultGlobalLibraryDirectories, getSysRoot(), ClangCppCompiler.class);
	}

	public BuildMessages compile(AbsoluteFilePath source) {
		final BuildMessages buildMessages = new BuildMessages();
		final AbsoluteFilePath object = getObjectPath(source);

		if (!object.getDirectory().create()) {
			object.tryDelete();
		}

		final Program.Parameters parameters = prepareParameters(source, object);

		final Program.EventHandler eventHandler = new GccOutputParser(buildMessages, source, isVerboseBuild());
		if (isVerboseBuild()) {
			System.out.println(compilerPath + " " + parameters.toString());
		}
		final Program compiler = new Program(compilerPath, parameters, new Program.Environment(), eventHandler);
		compiler.join();
		return buildMessages;
	}

	protected final Program.Parameters prepareParameters(AbsoluteFilePath sourcePath, AbsoluteFilePath objectPath) {
		final Program.Parameters parameters = new Program.Parameters();
		parameters.add("-c");
		if (getDebugInformation()) {
			parameters.add("-g");
		}
		switch (pic) {
			case NoPIC:
				parameters.add("-fno-pic");
				break;
			case LimitedLibraryPIC:
				parameters.add("-fpic");
				break;
			case UnlimitedLibraryPIC:
				parameters.add("-fPIC");
				break;
			case LimitedExecutablePIC:
				parameters.add("-fpie");
				break;
			case UnlimitedExecutablePIC:
				parameters.add("-fPIE");
				break;
		}
		if (getSysRoot() != null) {
			parameters.add("--sysroot=" + getSysRoot().getCanonicalForm());
		}
		switch (getOptimization()) {
			case NoOptimization:
				parameters.add("-O0");
				switch (getABI()) {
					case ARM_Linux_SoftEABI_Android:
					case ARM_Linux_SoftEABI_V5T:
						parameters.add("-marm");
						break;
					case ARM_Linux_SoftEABI_AndroidV7A:
					case ARM_Linux_HardEABI_V7A:
						parameters.add("-marm");
						break;
				}
				break;
			case MinSizeOptimization:
				parameters.add("-Os");
				switch (getABI()) {
					case ARM_Linux_SoftEABI_Android:
					case ARM_Linux_SoftEABI_V5T:
					case ARM_Linux_SoftEABI_AndroidV7A:
					case ARM_Linux_HardEABI_V7A:
						parameters.add("-mthumb");
						break;
				}
				break;
			case SpeedOptimization:
				parameters.add("-O2");
				switch (getABI()) {
					case ARM_Linux_SoftEABI_Android:
					case ARM_Linux_SoftEABI_V5T:
						parameters.add("-marm");
						break;
					case ARM_Linux_SoftEABI_AndroidV7A:
					case ARM_Linux_HardEABI_V7A:
						parameters.add("-mthumb");
						break;
				}
				break;
			case MaxSpeedOptimization:
				parameters.add("-O3");
				switch (getABI()) {
					case ARM_Linux_SoftEABI_Android:
					case ARM_Linux_SoftEABI_V5T:
						parameters.add("-marm");
						break;
					case ARM_Linux_SoftEABI_AndroidV7A:
					case ARM_Linux_HardEABI_V7A:
						parameters.add("-mthumb");
						break;
				}
				break;
		}
		if (getGlobalOptimization()) {
			parameters.add("-flto");
		}
		for (Map.Entry<String, String> macro: macros.entrySet()) {
			parameters.add("-D" + macro.getKey() + "=" + macro.getValue());
		}
		parameters.add("-Wa,--noexecstack");
		switch (runtimeLibrary) {
			case NoRuntimeLibrary:
				parameters.add("-nostdlib", "-fno-stack-protector"); break;
			case DynamicRuntimeLibrary:
				if (!getABI().getOperatingSystem().equals(OperatingSystem.MacOSX)) {
					parameters.add("-shared-libgcc");
				}
				break;
			case StaticRuntimeLibrary:
				if (!getABI().getOperatingSystem().equals(OperatingSystem.MacOSX)) {
					parameters.add("-static-libgcc");
				}
				break;
		}
		if (rttiEnabled != null) {
			parameters.add(rttiEnabled ? "-frtti" : "-fno-rtti");
		}
		if (exceptions != null) {
			switch (exceptions) {
				case NoExceptions:
					parameters.add("-fno-exceptions", "-fno-unwind-tables");
					break;
				case DisabledExceptions:
					parameters.add("-fno-exceptions", "-funwind-tables");
					break;
				case SyncronousExceptions:
					parameters.add("-fexceptions", "-funwind-tables");
					break;
				case AsyncronousExceptions:
					parameters.add("-fexceptions", "-funwind-tables", "-fasynchronous-unwind-tables", "-fnon-call-exceptions");
					break;
			}
		}
		switch (getABI()) {
			case ARM_Linux_SoftEABI_Android:
				parameters.add("-mfloat-abi=soft");
				parameters.add("-mtune=arm1136j-s");
				parameters.add("-march=armv5te");
				parameters.add("-mthumb-interwork");
				break;
			case ARM_Linux_SoftEABI_V5T:
				parameters.add("-mfloat-abi=soft");
				parameters.add("-mtune=arm1136j-s");
				parameters.add("-march=armv5t");
				parameters.add("-mthumb-interwork");
				break;
			case ARM_Linux_SoftEABI_AndroidV7A:
				parameters.add("-mfloat-abi=softfp");
				parameters.add("-mtune=cortex-a9");
				parameters.add("-march=armv7-a");
				parameters.add("-mfpu=vfpv3-d16");
				parameters.add("-mthumb-interwork");
				break;
			case ARM_Linux_HardEABI_V7A:
				parameters.add("-mfloat-abi=hard");
				parameters.add("-mtune=cortex-a9");
				parameters.add("-march=armv7-a");
				parameters.add("-mfpu=vfpv3-d16");
				parameters.add("-mthumb-interwork");
				break;
			case X64_Linux_SystemV_Default:
				parameters.add("-m64");
				parameters.add("-mtune=corei7");
				break;
			case X64_MacOSX_SystemV_Default:
				parameters.add("-m64");
				parameters.add("-mssse3");
				parameters.add("-mtune=corei7");
				break;
			case X86_Linux_Pic_Android:
				parameters.add("-m32");
				parameters.add("-msse3");
				parameters.add("-mtune=atom");
				parameters.add("-mfpmath=sse");
				break;
			case X86_Linux_Pic_i586:
				parameters.add("-m32");
				parameters.add("-mtune=corei7");
				parameters.add("-march=i586");
				break;
			case X86_MacOSX_Pic_Default:
				parameters.add("-m32");
				parameters.add("-msse3");
				parameters.add("-mtune=corei7");
				break;
			case MIPS_Linux_O32_Android:
				parameters.add("-march=mips32");
				parameters.add("-fno-inline-functions-called-once");
				parameters.add("-fgcse-after-reload");
				parameters.add("-frerun-cse-after-loop");
				parameters.add("-frename-registers");
				break;
			default:
				throw new Error(String.format("Unsupported ABI %s", getABI().toString()));
		}
		if (getABI().getOperatingSystem().equals(OperatingSystem.Linux)) {
			parameters.add("-ffunction-sections", "-fdata-sections");
		}
		parameters.add("-fomit-frame-pointer", "-fstrict-aliasing");
		for (final AbsoluteDirectoryPath includeDirectory : includeDirectories) {
			parameters.add("-I" + includeDirectory.getCanonicalForm());
		}
		for (final AbsoluteDirectoryPath globalIncludeDirectory : globalIncludeDirectories) {
			parameters.add("-I" + globalIncludeDirectory.getCanonicalForm());
		}
		parameters.add(sourcePath.getCanonicalForm());
		parameters.add("-o", objectPath.getCanonicalForm());
		return parameters;
	}

	@Override
	public final void addDefaultGlobalIncludeDirectories() {
		addGlobalIncludeDirectories(defaultIncludeDirectories);
	}

	@Override
	public CppCompiler copy() {
		return new ClangCppCompiler(this);
	}

	@Override
	public final Version getVersion() {
		return this.version;
	}

	private final AbsoluteFilePath compilerPath;
	private final List<AbsoluteDirectoryPath> defaultIncludeDirectories;
	private final Version version;
}
