/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.Linker;
import info.yeppp.ebuilda.internals.Program;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class GccCompatibleCompilerLinker extends AbstractLinker {

	public static ToolInfo verify(AbsoluteFilePath linkerPath) {
		return GnuCppCompiler.verify(linkerPath);
	}

	GccCompatibleCompilerLinker(AbsoluteFilePath linkerPath, ABI abi, Version version, Collection<AbsoluteDirectoryPath> defaultGlobalLibraryDirectories, AbsoluteDirectoryPath sysRoot, Class compiler) {
		super(abi);
		this.linkerPath = linkerPath;
		this.version = version;
		this.defaultGlobalLibraryDirectories = new ArrayList<AbsoluteDirectoryPath>(defaultGlobalLibraryDirectories);
		this.sysRoot = sysRoot;
		this.compiler = compiler;
	}

	public GccCompatibleCompilerLinker(GccCompatibleCompilerLinker prototype) {
		super(prototype);
		this.linkerPath = prototype.linkerPath;
		this.version = prototype.version;
		this.defaultGlobalLibraryDirectories = new ArrayList<AbsoluteDirectoryPath>(prototype.defaultGlobalLibraryDirectories);
		this.sysRoot = prototype.sysRoot;
		this.compiler = prototype.compiler;
	}

	public BuildMessages linkExecutable(AbsoluteFilePath executablePath, Collection<AbsoluteFilePath> objectPaths) {
		if (getBinariesDirectory() == null) {
			throw new IllegalStateException("Binaries directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();

		if (!executablePath.getDirectory().create()) {
			executablePath.tryDelete();
		}

		final Program.Parameters parameters = prepareParameters(executablePath, objectPaths, Target.Executable);
		final Program.Environment environment = prepareEnvironment();
		final Program.EventHandler eventHandler = new GccOutputParser(buildMessages, executablePath, isVerboseBuild());

		if (isVerboseBuild()) {
			System.out.println(linkerPath + " " + parameters.toString());
		}
		final Program linker = new Program(linkerPath, parameters, environment, eventHandler);
		linker.join();
		return buildMessages;
	}

	public BuildMessages linkDynamicLibrary(AbsoluteFilePath libraryPath, Collection<AbsoluteFilePath> objectPaths) {
		if (getBinariesDirectory() == null) {
			throw new IllegalStateException("Binaries directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();

		if (getABI().getOperatingSystem().equals(OperatingSystem.MacOSX)) {
			libraryPath = libraryPath.setExtension("dylib");
		} else {
			libraryPath = libraryPath.setExtension("so");
		}
		final String libraryName = "lib" + libraryPath.getFileName();
		libraryPath = new AbsoluteFilePath(libraryPath.getDirectory(), new RelativeFilePath(libraryName));
		if (!libraryPath.getDirectory().create()) {
			libraryPath.tryDelete();
		}

		final Program.Parameters parameters = prepareParameters(libraryPath, objectPaths, Target.DynamicLibrary);
		final Program.Environment environment = prepareEnvironment();
		final Program.EventHandler eventHandler = new GccOutputParser(buildMessages, libraryPath, isVerboseBuild());

		if (isVerboseBuild()) {
			System.out.println(linkerPath + " " + parameters.toString());
		}
		final Program linker = new Program(linkerPath, parameters, environment, eventHandler);
		linker.join();
		return buildMessages;
	}

	private Program.Parameters prepareParameters(AbsoluteFilePath outputPath, Collection<AbsoluteFilePath> objectPaths, Target target) {
		final Program.Parameters parameters = new Program.Parameters();
		switch (getPIC()) {
			case NoPIC:
				parameters.add("-fno-pic"); break;
			case LimitedExecutablePIC:
				parameters.add("-fpie");
				if (target.equals(Target.Executable)) {
					parameters.add("-pie");
				}
				break;
			case LimitedLibraryPIC:
				parameters.add("-fpic"); break;
			case UnlimitedExecutablePIC:
				parameters.add("-fPIE");
				if (target.equals(Target.Executable)) {
					parameters.add("-pie");
				}
				break;
			case UnlimitedLibraryPIC:
				parameters.add("-fPIC"); break;
		}
		if (getSysRoot() != null) {
			parameters.add("--sysroot=" + getSysRoot().getCanonicalForm());
		}
		if (getGlobalOptimization()) {
			parameters.add("-flto");
		}
		if (target.equals(Target.DynamicLibrary)) {
			if (getABI().getOperatingSystem().equals(OperatingSystem.MacOSX)) {
				parameters.add("-dynamiclib");
			} else {
				parameters.add("-shared");
			}
		}
		if (!getRuntimeLibraryUse()) {
			parameters.add("-nostdlib");
		}
		switch (getABI()) {
			case X86_Linux_Pic_i586:
			case X86_MacOSX_Pic_Default:
			case X86_NaCl_NaCl_i686:
				parameters.add("-m32"); break;
			case X64_Linux_SystemV_Default:
			case X64_MacOSX_SystemV_Default:
			case X64_NaCl_NaCl_Default:
				parameters.add("-m64"); break;
			case X64_Linux_KNC_Default:
				parameters.add(compiler.equals(IntelCppCompiler.class) ? "-mmic" : "-m64"); break;
		}
		if (getABI().getOperatingSystem().equals(OperatingSystem.Linux)) {
			parameters.add("-Wl,-z,noexecstack");
			parameters.add("-Wl,--no-undefined");
			parameters.add("-Wl,--gc-sections");
		}
		if (linkerPath.getMachine().getOperatingSystem().equals(OperatingSystem.Windows)) {
			for (final AbsoluteDirectoryPath globalLibraryDirectory : globalLibraryDirectories) {
				parameters.add("-L" + globalLibraryDirectory.getCanonicalForm());
			}
		}
		for (final AbsoluteDirectoryPath libraryDirectory : this.libraryDirectories) {
			parameters.add("-L" + libraryDirectory.getCanonicalForm());
		}
		parameters.add("-o", outputPath.getCanonicalForm());
		for (final AbsoluteFilePath objectPath : objectPaths) {
			parameters.add(objectPath.getCanonicalForm());
		}
		if (!staticLibraries.isEmpty()) {
			if (!getABI().getOperatingSystem().equals(OperatingSystem.MacOSX)) {
				parameters.add("-Wl,-Bstatic");
			}
			for (final String library : staticLibraries) {
				parameters.add("-l" + library);
			}
		}
		if (!dynamicLibraries.isEmpty()) {
			if (!getABI().getOperatingSystem().equals(OperatingSystem.MacOSX)) {
				parameters.add("-Wl,-Bdynamic");
			}
			for (final String library : dynamicLibraries) {
				parameters.add("-l" + library);
			}
		}
		return parameters;
	}

	private Program.Environment prepareEnvironment() {
		final Program.Environment environment = new Program.Environment();
		if (!linkerPath.getMachine().getOperatingSystem().equals(OperatingSystem.Windows)) {
			for (final AbsoluteDirectoryPath globalLibraryDirectory : globalLibraryDirectories) {
				environment.prependPathVariable("LIB", globalLibraryDirectory.getCanonicalForm());
			}
		}
		return environment;
	}

	@Override
	public void addDefaultGlobalLibraryDirectories() {
	}

	@Override
	public Linker copy() {
		return new GccCompatibleCompilerLinker(this);
	}

	public final AbsoluteDirectoryPath getSysRoot() {
		return sysRoot;
	}

	public final void setSysRoot(AbsoluteDirectoryPath sysRoot) {
		this.sysRoot = sysRoot;
	}

	private final AbsoluteFilePath linkerPath;
	private List<AbsoluteDirectoryPath> defaultGlobalLibraryDirectories;
	private final Version version;
	private AbsoluteDirectoryPath sysRoot;
	private final Class compiler;
}
