/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeDirectoryPath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.Versioned;

import java.util.*;
import java.util.regex.Pattern;

public class GccToolchain implements Versioned {

	private static final Pattern prefixPattern = Pattern.compile("(?:(:?(:?\\w+|\\d+(:?\\.\\d+)*))\\-)*");
	private static final Pattern suffixPattern = Pattern.compile("(?:(:?\\-(:?\\w+|\\d+(:?\\.\\d+)*)))*");

	private static Map<ABI, List<ToolsetInfo>> initLocalBinaries() {
		final String executableExtension = Machine.getLocal().getOperatingSystem().equals(OperatingSystem.Windows) ? ".exe" : "";
		final List<ToolsetInfo> toolsetCandidates = new ArrayList<ToolsetInfo>();
		for (final AbsoluteDirectoryPath path : Machine.getLocal().getStandardBinaryDirectories()) {
			toolsetCandidates.addAll(ToolsetInfo.enumerate(path, prefixPattern, suffixPattern, new String[]{"gcc", "g++"}));
		}
		final List<ToolsetInfo> toolsets = new ArrayList<ToolsetInfo>(toolsetCandidates.size());
		for (final ToolsetInfo toolsetCandidate : toolsetCandidates) {
			final RelativeFilePath cppCompilerName = new RelativeFilePath(toolsetCandidate.getPrefix() + "g++" + toolsetCandidate.getSuffix() + executableExtension);
			final AbsoluteFilePath cppCompilerPath = new AbsoluteFilePath(toolsetCandidate.getPath(), cppCompilerName);
			final ToolInfo cppCompilerInfo = GnuCppCompiler.verify(cppCompilerPath);
			if (cppCompilerInfo.isValid() && (cppCompilerInfo.getDefaultABI() != null) && (cppCompilerInfo.getSupportedABI() != null)) {
				final ABI defaultABI = cppCompilerInfo.getDefaultABI();
				final ABI[] supportedABI = cppCompilerInfo.getSupportedABI();
				final Version version = cppCompilerInfo.getVersion();
				toolsets.add(new ToolsetInfo(toolsetCandidate.getPath(), toolsetCandidate.getPrefix(), toolsetCandidate.getSuffix(),
						defaultABI, supportedABI, version));
			}
		}
		final HashMap<ABI, List<ToolsetInfo>> abiMap = new HashMap<ABI, List<ToolsetInfo>>(ABI.values().length);
		for (final ABI abi : ABI.values()) {
			abiMap.put(abi, new LinkedList<ToolsetInfo>());
		}
		for (final ToolsetInfo toolset : toolsets) {
			for (final ABI abi : toolset.getSupportedABI()) {
				abiMap.get(abi).add(toolset);
			}
		}
		return abiMap;
	}

	private static Map<ABI, List<ToolsetInfo>> localBinaries = GccToolchain.initLocalBinaries();

	public static VersionList<GccToolchain> enumerate(Machine machine, ABI abi) {
		final VersionList<GccToolchain> toolchains = new VersionList<GccToolchain>();
		for (final ToolsetInfo toolsetInfo : localBinaries.get(abi)) {
			final GccToolchain toolchain = new GccToolchain(toolsetInfo.getPath(), toolsetInfo.getPrefix(), toolsetInfo.getSuffix(), abi, toolsetInfo.getVersion(), null);
			toolchains.add(toolchain);
		}
		return toolchains;
	}

	public GccToolchain(AbsoluteDirectoryPath root, String prefix, String suffix, ABI abi, Version version, AbsoluteDirectoryPath sysRoot) {
		final AbsoluteFilePath cppCompilerPath = new AbsoluteFilePath(root, new RelativeFilePath(prefix + "g++" + suffix));
		final AbsoluteFilePath cCompilerPath = new AbsoluteFilePath(root, new RelativeFilePath(prefix + "gcc" + suffix));
		this.abi = abi;
		this.version = version;
		this.sysRoot = sysRoot;
		final List<AbsoluteDirectoryPath> defaultIncludeDirectories = new LinkedList<AbsoluteDirectoryPath>();
		if (sysRoot != null) {
			defaultIncludeDirectories.add(new AbsoluteDirectoryPath(sysRoot, new RelativeDirectoryPath("usr/include")));
		}
		this.cppCompiler = new GnuCppCompiler(cppCompilerPath, abi, version, defaultIncludeDirectories, sysRoot);
		this.cCompiler = new GnuCCompiler(cCompilerPath, abi, version, defaultIncludeDirectories, sysRoot);
	}

	public GnuCppCompiler getCppCompiler() {
		return new GnuCppCompiler(this.cppCompiler);
	}

	public GnuCCompiler getCCompiler() {
		return new GnuCCompiler(this.cCompiler);
	}

	public ABI getABI() {
		return this.abi;
	}

	@Override
	public Version getVersion() {
		return this.version;
	}

	public AbsoluteDirectoryPath getSysRoot() {
		return this.sysRoot;
	}

	@Override
	public String toString() {
		return String.format("GCC version %s for %s", getVersion().toString(), getABI().toString());
	}

	final private ABI abi;
	final private Version version;
	final private AbsoluteDirectoryPath sysRoot;
	final private GnuCppCompiler cppCompiler;
	final private GnuCCompiler cCompiler;
}
