/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.Versioned;

import java.util.*;
import java.util.regex.Pattern;

public class GnuBinutils implements Versioned {

	private static final Pattern prefixPattern = Pattern.compile("(?:(:?(:?\\w+|\\d+(:?\\.\\d+)*))\\-)*");
	private static final Pattern suffixPattern = Pattern.compile("(?:(:?\\-(:?\\w+|\\d+(:?\\.\\d+)*)))*");

	private static Map<ABI, List<ToolsetInfo>> initLocalBinaries() {
		final String executableExtension = Machine.getLocal().getOperatingSystem().equals(OperatingSystem.Windows) ? ".exe" : "";
		final List<ToolsetInfo> toolsetCandidates = new ArrayList<ToolsetInfo>();
		if (!Machine.getLocal().getOperatingSystem().equals(OperatingSystem.MacOSX)) {
			for (final AbsoluteDirectoryPath path : Machine.getLocal().getStandardBinaryDirectories()) {
				toolsetCandidates.addAll(ToolsetInfo.enumerate(path, prefixPattern, suffixPattern, new String[]{"as", "ar", "strip", "objcopy"}));
			}
		}
		final List<ToolsetInfo> toolsets = new ArrayList<ToolsetInfo>(toolsetCandidates.size());
		for (final ToolsetInfo toolsetCandidate : toolsetCandidates) {
			final RelativeFilePath assemblerName = new RelativeFilePath(toolsetCandidate.getPrefix() + "as" + toolsetCandidate.getSuffix() + executableExtension);
			final AbsoluteFilePath assemblerPath = new AbsoluteFilePath(toolsetCandidate.getPath(), assemblerName);
			final ToolInfo assemblerInfo = GnuAssembler.verify(assemblerPath);
			if (assemblerInfo.isValid() && (assemblerInfo.getDefaultABI() != null) && (assemblerInfo.getSupportedABI() != null)) {
				final RelativeFilePath stripName = new RelativeFilePath(toolsetCandidate.getPrefix() + "strip" + toolsetCandidate.getSuffix() + executableExtension);
				final AbsoluteFilePath stripPath = new AbsoluteFilePath(toolsetCandidate.getPath(), stripName);
				final ToolInfo stripInfo = GnuStrip.verify(stripPath);

				if (stripInfo.isValid()) {
					final ABI defaultABI = assemblerInfo.getDefaultABI();
					final ABI[] supportedABI = assemblerInfo.getSupportedABI();
					final Version version = assemblerInfo.getVersion();
					toolsets.add(new ToolsetInfo(toolsetCandidate.getPath(), toolsetCandidate.getPrefix(), toolsetCandidate.getSuffix(),
							defaultABI, supportedABI, version));
				}
			}
		}
		final HashMap<ABI, List<ToolsetInfo>> abiMap = new HashMap<ABI, List<ToolsetInfo>>(ABI.values().length);
		for (final ABI abi : ABI.values()) {
			abiMap.put(abi, new LinkedList<ToolsetInfo>());
		}
		for (final ToolsetInfo toolset : toolsets) {
			for (final ABI abi : toolset.getSupportedABI()) {
				abiMap.get(abi).add(toolset);
			}
		}
		return abiMap;
	}

	private static Map<ABI, List<ToolsetInfo>> localBinaries = GnuBinutils.initLocalBinaries();

	public static VersionList<GnuBinutils> enumerate(Machine machine, ABI abi) {
		final VersionList<GnuBinutils> toolchains = new VersionList<GnuBinutils>();
		for (final ToolsetInfo toolsetInfo : localBinaries.get(abi)) {
			final GnuBinutils toolchain = new GnuBinutils(toolsetInfo.getPath(), toolsetInfo.getPrefix(), toolsetInfo.getSuffix(), abi, toolsetInfo.getVersion(), null);
			toolchains.add(toolchain);
		}
		return toolchains;
	}

	public GnuBinutils(AbsoluteDirectoryPath root, String prefix, String suffix, ABI abi, Version version, AbsoluteDirectoryPath sysRoot) {
		final AbsoluteFilePath stripPath = new AbsoluteFilePath(root, new RelativeFilePath(prefix + "strip" + suffix));
		final AbsoluteFilePath objCopyPath = new AbsoluteFilePath(root, new RelativeFilePath(prefix + "objcopy" + suffix));
		final AbsoluteFilePath assemblerPath = new AbsoluteFilePath(root, new RelativeFilePath(prefix + "as" + suffix));
		this.abi = abi;
		this.version = version;
		this.sysRoot = sysRoot;
		this.strip = new GnuStrip(stripPath, abi);
		this.objCopy = new GnuObjCopy(objCopyPath, abi);
		this.assembler = new GnuAssembler(assemblerPath, abi);
	}

	public GnuAssembler getAssembler() {
		return new GnuAssembler(this.assembler);
	}

	public GnuStrip getStrip() {
		return new GnuStrip(this.strip);
	}

	public GnuObjCopy getObjCopy() {
		return new GnuObjCopy(this.objCopy);
	}

	public ABI getABI() {
		return this.abi;
	}

	public Version getVersion() {
		return this.version;
	}

	public AbsoluteDirectoryPath getSysRoot() {
		return this.sysRoot;
	}

	@Override
	public String toString() {
		return String.format("GNU binutils version %s for %s", getVersion().toString(), getABI().toString());
	}

	final private ABI abi;
	final private Version version;
	final private AbsoluteDirectoryPath sysRoot;
	final private GnuAssembler assembler;
	final private GnuStrip strip;
	final private GnuObjCopy objCopy;
}
