/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.internals.Program;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class GnuStrip {

	/* Examples of first output line:
	 * "GNU strip (GNU Binutils) 2.22.52.20120302"
	 * "GNU strip (GNU Binutils for Ubuntu) 2.23.2"
	 * "GNU strip (GNU Binutils for Debian) 2.22"
	 * "GNU strip (GNU Binutils; SUSE Linux Enterprise PTF) 2.21.1"
	 * "GNU strip (Native Client SDK [38db877804d688ceacbcfafd85fb205cd47296fb]) 2.23.1.20121210"
	 * "GNU strip version 2.20.51.0.2-5.36.el6 20100205"
	 * "GNU strip version 2.21.53.0.1-2.fc16 20110716"
	 */
	private static final Pattern debianVersionPattern = Pattern.compile("GNU strip \\(.+\\) (\\d+(?:\\.\\d+)+)\\b\\.*");
	private static final Pattern redhatVersionPattern = Pattern.compile("GNU strip version (\\d+(?:\\.\\d+)+)\\-.+");

	public static ToolInfo verify(AbsoluteFilePath stripPath) {
		final Program.Environment environment = new Program.Environment();
		final List<String> versionStrings = Program.getStdoutLines(stripPath, new Program.Parameters("--version"), environment, 1);
		if (versionStrings.isEmpty()) {
			return new ToolInfo(stripPath);
		} else {
			final String versionString = versionStrings.get(0);
			final Matcher versionMatcher = debianVersionPattern.matcher(versionString).matches() ?
					debianVersionPattern.matcher(versionString) : redhatVersionPattern.matcher(versionString);
			if (versionMatcher.matches()) {
				final Version version = new Version(versionMatcher.group(1));
				final List<String> targetStrings = Program.getStdoutLines(stripPath, new Program.Parameters("--info"), environment, 1024);
				final List<ABI> supportedABI = new ArrayList<ABI>(ABI.values().length);
				ABI defaultABI = null;
				for (final String targetString : targetStrings) {
					if (!targetString.isEmpty()) {
						if (!Character.isWhitespace(targetString.charAt(0))) {
							if (targetString.equals("elf64-x86-64")) {
								supportedABI.add(ABI.X64_Linux_SystemV_Default);
							} else if (targetString.equals("elf32-i386")) {
								supportedABI.add(ABI.X86_Linux_Pic_i586);
								supportedABI.add(ABI.X86_Linux_Pic_Android);
							} else if (targetString.equals("elf64-k1om")) {
								supportedABI.add(ABI.X64_Linux_KNC_Default);
							} else if (targetString.equals("elf32-littlearm")) {
								supportedABI.add(ABI.ARM_Linux_SoftEABI_V5T);
								supportedABI.add(ABI.ARM_Linux_HardEABI_V7A);
								supportedABI.add(ABI.ARM_Linux_SoftEABI_Android);
								supportedABI.add(ABI.ARM_Linux_SoftEABI_AndroidV7A);
							} else if (targetString.equals("elf32-tradlittlemips")) {
								supportedABI.add(ABI.MIPS_Linux_O32_Android);
							} else if (targetString.equals("elf32-littlearm-nacl")) {
								supportedABI.add(ABI.ARM_NaCl_NaCl_Default);
							} else if (targetString.equals("elf64-nacl")) {
								supportedABI.add(ABI.X64_NaCl_NaCl_Default);
							} else if (targetString.equals("elf32-nacl")) {
								supportedABI.add(ABI.X86_NaCl_NaCl_i686);
							}
						}
					}
				}
				return new ToolInfo(stripPath, null, ABI.values(), version);
			} else {
				return new ToolInfo(stripPath);
			}
		}
	}

	public GnuStrip(AbsoluteFilePath stripPath, ABI abi) {
		this.stripPath = stripPath;
		this.abi = abi;
		this.verboseBuild = false;
	}

	public GnuStrip(GnuStrip prototype) {
		this.stripPath = prototype.stripPath;
		this.abi = prototype.abi;
		this.verboseBuild = prototype.verboseBuild;
	}

	public BuildMessages strip(AbsoluteFilePath binary) {
		final BuildMessages buildMessages = new BuildMessages();
		final Program.Parameters parameters = new Program.Parameters();
		parameters.add("--strip-debug");
		parameters.add("--strip-unneeded");
		parameters.add("--remove-section=.comment");
		parameters.add(binary.getCanonicalForm());
		final Program.Environment environment = new Program.Environment();
		final Program.EventHandler eventHandler = new PassThroughOutputParser(buildMessages, binary);
		if (isVerboseBuild()) {
			System.out.println(stripPath + " " + parameters.toString());
		}
		final Program strip = new Program(stripPath, parameters, environment, eventHandler);
		strip.join();
		return buildMessages;
	}

	public BuildMessages removeSection(AbsoluteFilePath binary, String section) {
		final BuildMessages buildMessages = new BuildMessages();
		final Program.Parameters parameters = new Program.Parameters();
		parameters.add("--remove-section=" + section);
		parameters.add(binary.getCanonicalForm());
		final Program.Environment environment = new Program.Environment();
		final Program.EventHandler eventHandler = new PassThroughOutputParser(buildMessages, binary);
		if (isVerboseBuild()) {
			System.out.println(stripPath + " " + parameters.toString());
		}
		final Program strip = new Program(stripPath, parameters, environment, eventHandler);
		strip.join();
		return buildMessages;
	}

	public BuildMessages extractDebugInformation(AbsoluteFilePath binary, AbsoluteFilePath debugInformation) {
		final BuildMessages buildMessages = new BuildMessages();
		final Program.Parameters parameters = new Program.Parameters();
		parameters.add("--only-keep-debug");
		parameters.add(binary.getCanonicalForm());
		parameters.add("-o", debugInformation.getCanonicalForm());
		final Program.Environment environment = new Program.Environment();
		final Program.EventHandler eventHandler = new PassThroughOutputParser(buildMessages, binary);
		if (isVerboseBuild()) {
			System.out.println(stripPath + " " + parameters.toString());
		}
		final Program strip = new Program(stripPath, parameters, environment, eventHandler);
		strip.join();
		return buildMessages;
	}

	public void setVerboseBuild(boolean verboseBuild) {
		this.verboseBuild = verboseBuild;
	}

	public boolean isVerboseBuild() {
		return this.verboseBuild;
	}

	public ABI getABI() {
		return abi;
	}

	private final AbsoluteFilePath stripPath;
	private final ABI abi;
	private boolean verboseBuild;
}
