/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeDirectoryPath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.Versioned;

import java.util.*;
import java.util.regex.Pattern;

public class IntelCppToolchain implements Versioned {

	private static final Pattern emptyPattern = Pattern.compile("");

	private static Map<ABI, List<ToolsetInfo>> initLocalBinaries() {
		final String executableExtension = Machine.getLocal().getOperatingSystem().equals(OperatingSystem.Windows) ? ".exe" : "";
		final List<ToolsetInfo> toolsetCandidates = new ArrayList<ToolsetInfo>();
		for (final AbsoluteDirectoryPath path : Machine.getLocal().getStandardBinaryDirectories()) {
			toolsetCandidates.addAll(ToolsetInfo.enumerate(path, emptyPattern, emptyPattern, new String[]{"icc", "icpc"}));
		}
		if (Machine.getLocal().getOperatingSystem().equals(OperatingSystem.Linux)) {
			final AbsoluteDirectoryPath expectedDirectory = new AbsoluteDirectoryPath(Machine.getLocal(), "/opt/intel/composerxe/bin");
			toolsetCandidates.add(new ToolsetInfo(expectedDirectory, "", ""));
		}
		final List<ToolsetInfo> toolsets = new ArrayList<ToolsetInfo>(toolsetCandidates.size());
		for (final ToolsetInfo toolsetCandidate : toolsetCandidates) {
			final RelativeFilePath cppCompilerName = new RelativeFilePath(toolsetCandidate.getPrefix() + "icpc" + toolsetCandidate.getSuffix() + executableExtension);
			final AbsoluteFilePath cppCompilerPath = new AbsoluteFilePath(toolsetCandidate.getPath(), cppCompilerName);
			final ToolInfo cppCompilerInfo = IntelCppCompiler.verify(cppCompilerPath);
			if (cppCompilerInfo.isValid() && (cppCompilerInfo.getDefaultABI() != null) && (cppCompilerInfo.getSupportedABI() != null)) {
				final ABI defaultABI = cppCompilerInfo.getDefaultABI();
				final ABI[] supportedABI = cppCompilerInfo.getSupportedABI();
				final Version version = cppCompilerInfo.getVersion();
				toolsets.add(new ToolsetInfo(toolsetCandidate.getPath(), toolsetCandidate.getPrefix(), toolsetCandidate.getSuffix(),
						defaultABI, supportedABI, version));
			}
		}
		final HashMap<ABI, List<ToolsetInfo>> abiMap = new HashMap<ABI, List<ToolsetInfo>>(ABI.values().length);
		for (final ABI abi : ABI.values()) {
			abiMap.put(abi, new LinkedList<ToolsetInfo>());
		}
		for (final ToolsetInfo toolset : toolsets) {
			for (final ABI abi : toolset.getSupportedABI()) {
				abiMap.get(abi).add(toolset);
			}
		}
		return abiMap;
	}

	private static Map<ABI, List<ToolsetInfo>> localBinaries = IntelCppToolchain.initLocalBinaries();

	public static VersionList<IntelCppToolchain> enumerate(Machine machine, ABI abi) {
		final VersionList<IntelCppToolchain> toolchains = new VersionList<IntelCppToolchain>();
		for (final ToolsetInfo toolsetInfo : localBinaries.get(abi)) {
			final IntelCppToolchain toolchain = new IntelCppToolchain(toolsetInfo.getPath(), toolsetInfo.getPrefix(), toolsetInfo.getSuffix(), abi, toolsetInfo.getVersion(), null);
			toolchains.add(toolchain);
		}
		return toolchains;
	}

	public IntelCppToolchain(AbsoluteDirectoryPath root, String prefix, String suffix, ABI abi, Version version, AbsoluteDirectoryPath sysRoot) {
		final AbsoluteFilePath cppCompilerPath = new AbsoluteFilePath(root, new RelativeFilePath(prefix + "icpc" + suffix));
		final AbsoluteFilePath cCompilerPath = new AbsoluteFilePath(root, new RelativeFilePath(prefix + "icc" + suffix));
		this.abi = abi;
		this.version = version;
		this.sysRoot = sysRoot;
		final List<AbsoluteDirectoryPath> defaultIncludeDirectories = new LinkedList<AbsoluteDirectoryPath>();
		if (sysRoot != null) {
			defaultIncludeDirectories.add(new AbsoluteDirectoryPath(sysRoot, new RelativeDirectoryPath("usr/include")));
		}
		this.cppCompiler = new IntelCppCompiler(cppCompilerPath, abi, version, defaultIncludeDirectories, sysRoot);
		this.cCompiler = new IntelCCompiler(cCompilerPath, abi, version, defaultIncludeDirectories, sysRoot);
	}

	public IntelCppCompiler getCppCompiler() {
		return new IntelCppCompiler(this.cppCompiler);
	}

	public IntelCCompiler getCCompiler() {
		return new IntelCCompiler(this.cCompiler);
	}

	public ABI getABI() {
		return this.abi;
	}

	@Override
	public Version getVersion() {
		return this.version;
	}

	public AbsoluteDirectoryPath getSysRoot() {
		return this.sysRoot;
	}

	@Override
	public String toString() {
		return String.format("Intel C/C++ toolchain %s for %s", getVersion().toString(), getABI().toString());
	}

	final private ABI abi;
	final private Version version;
	final private AbsoluteDirectoryPath sysRoot;
	final private IntelCppCompiler cppCompiler;
	final private IntelCCompiler cCompiler;
}
