/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.Versioned;
import info.yeppp.ebuilda.internals.Program;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class JavacJavaCompiler extends AbstractJavaCompiler implements Versioned {

	private static final Pattern versionRegex = Pattern.compile("javac (\\d+(?:\\.\\d+)+_\\d+)");

	public static ToolInfo verify(AbsoluteFilePath compilerPath) {
		Program.Parameters parameters = new Program.Parameters("-version");
		Program.Environment environment = new Program.Environment();
		final List<String> outputStrings = Program.getStderrLines(compilerPath, parameters, environment, 1);
		if (outputStrings.isEmpty()) {
			return new ToolInfo(compilerPath);
		} else {
			final String versionString = outputStrings.get(0);
			final Matcher versionMatcher = versionRegex.matcher(versionString);
			if (versionMatcher.matches()) {
				final Version version = new Version(versionMatcher.group(1).replace('_', '.'));
				return new ToolInfo(compilerPath, null, null, version);
			} else {
				return new ToolInfo(compilerPath);
			}
		}
	}

	JavacJavaCompiler(AbsoluteFilePath compilerPath, Version version) {
		super();
		this.compilerPath = compilerPath;
		this.version = version;
	}

	public JavacJavaCompiler(JavacJavaCompiler prototype) {
		super(prototype);
		this.compilerPath = prototype.compilerPath;
		this.version = prototype.version;
	}

	public BuildMessages compile(AbsoluteFilePath source) {
		final BuildMessages buildMessages = new BuildMessages();
		final AbsoluteFilePath object = getObjectPath(source);

		if (!object.getDirectory().create()) {
			object.tryDelete();
		}

		final Program.Parameters parameters = prepareParameters(source, object);

		final Program.EventHandler eventHandler = new GccOutputParser(buildMessages, source, isVerboseBuild());
		if (isVerboseBuild()) {
			System.out.println(compilerPath + " " + parameters.toString());
		}
		final Program compiler = new Program(compilerPath, parameters, new Program.Environment(), eventHandler);
		compiler.join();
		return buildMessages;
	}

	protected final Program.Parameters prepareParameters(AbsoluteFilePath sourcePath, AbsoluteFilePath objectPath) {
		final Program.Parameters parameters = new Program.Parameters();
		return parameters;
	}

	@Override
	public JavacJavaCompiler copy() {
		return new JavacJavaCompiler(this);
	}

	@Override
	public final Version getVersion() {
		return this.version;
	}

	private final AbsoluteFilePath compilerPath;
	private final Version version;

}
