/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.PathFormatException;
import info.yeppp.ebuilda.generic.Linker;
import info.yeppp.ebuilda.internals.Program;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MicrosoftLinker extends AbstractLinker {

	private static class OutputParser implements Program.EventHandler {
		OutputParser(BuildMessages buildMessages, AbsoluteFilePath libraryPath, Collection<AbsoluteFilePath> objectsPaths, boolean verboseBuild, boolean codeGeneration) {
			this.buildMessages = buildMessages;
			this.exportLibraryPath = libraryPath == null ? null : libraryPath.setExtension("exp");
			this.importLibraryPath = libraryPath == null ? null : libraryPath.setExtension("lib");
			this.objectsPaths = objectsPaths;
			this.verboseBuild = verboseBuild;
			this.expectGeneratingCodeMessage = codeGeneration;
		}

		private final BuildMessages buildMessages;
		private final AbsoluteFilePath exportLibraryPath;
		private final AbsoluteFilePath importLibraryPath;
		private final Collection<AbsoluteFilePath> objectsPaths;
		private final boolean verboseBuild;
		private boolean isFirstOutputLine = true;
		private boolean expectGeneratingCodeMessage;
		private boolean expectFinishGeneratingCodeMessage = false;

		private static final Pattern outputRegex = Pattern.compile("(.+) : ((?:fatal )?error|warning) LNK(\\d+): (.+)");
//		private static final Pattern errorPattern = Pattern.compile("(.+)\\((\\d+)\\)\\s*: error C(\\d+): (.+)");
//		private static final Pattern fatalErrorPattern = Pattern.compile("(.+)\\((\\d+)\\)\\s*: fatal error C(\\d+): (.+)");
		private static final Pattern warningPattern = Pattern.compile("(.+)\\((\\d+)\\)\\s*: warning C(\\d+): (.+)");

		protected BuildMessages parseCommandLineOutput(List<String> lines) {
			final BuildMessages buildMessages = new BuildMessages();
			return buildMessages;
		}

		private static final BuildMessage.Type getMessageType(String type) {
			if (type.equals("fatal error")) {
				return BuildMessage.Type.FatalError;
			} else if (type.equals("error")) {
				return BuildMessage.Type.Error;
			} else if (type.equals("warning")) {
				return BuildMessage.Type.Warning;
			} else {
				throw new Error(String.format("Unknown linker message type %s", type));
			}
		}

		private final AbsoluteFilePath matchInputFile(String path) {
			if (path.equals("LINK")) {
				return null;
			} else {
				try {
					return new AbsoluteFilePath(Machine.getLocal(), path);
				} catch (PathFormatException e) {
					for (final AbsoluteFilePath objectPath : this.objectsPaths) {
						if (objectPath.getFileName().equals(path)) {
							return objectPath;
						}
					}
					return null;
				}
			}
		}

		@Override
		public void onStdOutLineWrite(String line) {
			if (isFirstOutputLine) {
				isFirstOutputLine = false;
				if (Pattern.matches("\\s*Creating library (.+)\\.lib and object \\1\\.exp", line)) {
					return;
				}
			}
			if (expectGeneratingCodeMessage) {
				if (line.equals("Generating code")) {
					expectGeneratingCodeMessage = false;
					expectFinishGeneratingCodeMessage = true;
					return;
				}
			} else if (expectFinishGeneratingCodeMessage) {
				if (line.equals("Finished generating code")) {
					expectFinishGeneratingCodeMessage = false;
					return;
				}
			}
			if (verboseBuild) {
				System.out.println(line);
			}
			final Matcher outputMatcher = outputRegex.matcher(line);
			if (outputMatcher.matches()) {
				final BuildMessage.Type type = getMessageType(outputMatcher.group(2));
				final AbsoluteFilePath path = matchInputFile(outputMatcher.group(1));
				final int id = Integer.parseInt(outputMatcher.group(3));
				final String message = outputMatcher.group(4);
				buildMessages.add(new BuildMessage(type, path, null, id, message));
			} else {
				final Matcher warningMatcher = warningPattern.matcher(line);
				if (warningMatcher.matches()) {
					final AbsoluteFilePath warningPath = new AbsoluteFilePath(Machine.getLocal(), warningMatcher.group(1));
					final int lineNumber = Integer.parseInt(warningMatcher.group(2));
					final int warningNumber = Integer.parseInt(warningMatcher.group(3));
					final String warningMessage = warningMatcher.group(4);
					buildMessages.add(new BuildMessage(BuildMessage.Type.Warning, warningPath, lineNumber, warningNumber, warningMessage));
				} else {
					final BuildMessage lastBuildMessage = buildMessages.getLastBuildMessage();
					if (lastBuildMessage == null) {
						throw new Error(String.format("Unknown compiler message %s", line));
					} else {
						lastBuildMessage.appendMessage("\r\n" + line);
					}
				}
			}
		}

		@Override
		public void onStdErrLineWrite(String line) {
			if (verboseBuild) {
				System.out.println(line);
			}
			throw new Error(String.format("Unknown compiler message %s", line));
		}

		@Override
		public void onError(String description) {
			if (verboseBuild) {
				System.err.println(description);
			}
			buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, null, null, null, description));
		}

		@Override
		public void onCompletion(Integer status) {
		}

	}

	protected MicrosoftLinker(AbsoluteFilePath linkerPath, ABI abi, Collection<AbsoluteDirectoryPath> additionalPathDirectories, Collection<AbsoluteDirectoryPath> defaultGlobalLibraryDirectories) {
		super(abi);
		this.linkerPath = linkerPath;
		this.additionalPathDirectories = new ArrayList<AbsoluteDirectoryPath>(additionalPathDirectories);
		this.defaultGlobalLibraryDirectories = new ArrayList<AbsoluteDirectoryPath>(defaultGlobalLibraryDirectories);
	}

	public MicrosoftLinker(MicrosoftLinker prototype) {
		super(prototype);
		this.linkerPath = prototype.linkerPath;
		this.additionalPathDirectories = new ArrayList<AbsoluteDirectoryPath>(prototype.additionalPathDirectories);
		this.defaultGlobalLibraryDirectories = new ArrayList<AbsoluteDirectoryPath>(prototype.defaultGlobalLibraryDirectories);
		this.largeAddressCompatible = prototype.largeAddressCompatible;
		this.depCompatible = prototype.depCompatible;
		this.safeSehCompatible = prototype.safeSehCompatible;
	}

	public BuildMessages linkExecutable(AbsoluteFilePath executablePath, Collection<AbsoluteFilePath> objectPaths) {
		if (getBinariesDirectory() == null) {
			throw new IllegalStateException("Binaries directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();

		executablePath = executablePath.setExtension("exe");
		if (!executablePath.getDirectory().create()) {
			executablePath.tryDelete();
		}

		final Program.Parameters parameters = prepareParameters(executablePath, objectPaths, Target.Executable);
		final Program.Environment environment = prepareEnvironment();
		final Program.EventHandler eventHandler = new OutputParser(buildMessages, null, objectPaths, isVerboseBuild(), getGlobalOptimization());

		if (isVerboseBuild()) {
			System.out.println(linkerPath.toString() + " " + parameters.toString());
		}
		final Program linker = new Program(linkerPath, parameters, environment, eventHandler);
		linker.join();
		return buildMessages;
	}

	@Override
	public BuildMessages linkDynamicLibrary(AbsoluteFilePath libraryPath, Collection<AbsoluteFilePath> objectPaths) {
		if (getBinariesDirectory() == null) {
			throw new IllegalStateException("Binaries directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();

		libraryPath = libraryPath.setExtension("dll");
		if (!libraryPath.getDirectory().create()) {
			libraryPath.tryDelete();
		}

		final Program.Parameters parameters = prepareParameters(libraryPath, objectPaths, Target.DynamicLibrary);
		final Program.Environment environment = prepareEnvironment();
		final Program.EventHandler eventHandler = new OutputParser(buildMessages, libraryPath, objectPaths, isVerboseBuild(), getGlobalOptimization());

		if (isVerboseBuild()) {
			System.out.println(linkerPath.toString() + " " + parameters.toString());
		}
		final Program linker = new Program(linkerPath, parameters, environment, eventHandler);
		linker.join();
		return buildMessages;
	}

	private Program.Parameters prepareParameters(AbsoluteFilePath outputPath, Collection<AbsoluteFilePath> objectPaths, Target target) {
		final Program.Parameters parameters = new Program.Parameters();
		parameters.add("/nologo", "/SUBSYSTEM:CONSOLE", "/INCREMENTAL:NO");
		if (target.equals(Target.DynamicLibrary)) {
			parameters.add("/DLL");
		}
		parameters.add(isDepCompatible() ? "/NXCOMPAT" : "/NXCOMPAT:NO");
		if (getABI().getArchitecture().equals(Architecture.X86)) {
			parameters.add(isSafeSehCompatible() ? "/SAFESEH" : "/SAFESEH:NO");
		}
		parameters.add(isLargeAddressCompatible() ? "/LARGEADDRESSAWARE" : "/LARGEADDRESSAWARE:NO");
		if (target.equals(Target.Executable)) {
			parameters.add("/TSAWARE");
		}
		if (!getRuntimeLibraryUse()) {
			parameters.add("/NODEFAULTLIB");
		}
		if (getDebugInformation()) {
			parameters.add("/DEBUG");
			final AbsoluteFilePath pdbPath = outputPath.replaceExtension("pdb");
			parameters.add("/PDB:" + pdbPath.getCanonicalForm());
			parameters.add("/PDBALTPATH:" + pdbPath.getFileName());
		}
		if (getGlobalOptimization()) {
			parameters.add("/LTCG");
		}
		for (final AbsoluteDirectoryPath libraryDirectory : libraryDirectories) {
			parameters.add("/LIBPATH:" + libraryDirectory.getCanonicalForm());
		}
		switch (getABI().getArchitecture()) {
			case X86:
				parameters.add("/MACHINE:X86"); break;
			case X64:
				parameters.add("/MACHINE:X64"); break;
			case IA64:
				parameters.add("/MACHINE:IA64"); break;
			case ARM:
				parameters.add("/MACHINE:ARM"); break;
			case MIPS:
				parameters.add("/MACHINE:MIPS"); break;
		}
		parameters.add("/RELEASE");
		parameters.add("/OUT:" + outputPath.getCanonicalForm());
		for (final AbsoluteFilePath objectPath : objectPaths) {
			parameters.add(objectPath.getCanonicalForm());
		}
		for (String library : staticLibraries) {
			parameters.add(library + ".lib");
		}
		for (String library : dynamicLibraries) {
			parameters.add(library + ".lib");
		}
		return parameters;
	}

	private Program.Environment prepareEnvironment() {
		final Program.Environment environment = new Program.Environment();
		for (final AbsoluteDirectoryPath additionalPathDirectory : additionalPathDirectories) {
			environment.prependPathVariable("PATH", additionalPathDirectory.getCanonicalForm());
		}
		for (final AbsoluteDirectoryPath globalLibraryDirectory : globalLibraryDirectories) {
			environment.prependPathVariable("LIB", globalLibraryDirectory.getCanonicalForm());
		}
		return environment;
	}

	@Override
	public void addDefaultGlobalLibraryDirectories() {
		this.addGlobalLibraryDirectories(this.defaultGlobalLibraryDirectories);
	}

	public final boolean isLargeAddressCompatible() {
		return this.largeAddressCompatible;
	}

	public final void setLargeAddressCompatible(boolean largeAddressCompatible) {
		if (!largeAddressCompatible){
			if (getABI().getArchitecture().is64bit()) {
				throw new IllegalArgumentException(String.format("Binaries for 64-bit architecture %s are always compatible with large addresses", getABI().getArchitecture().toString()));
			}
		}
		this.largeAddressCompatible = largeAddressCompatible;
	}

	public final boolean isDepCompatible() {
	    return this.depCompatible;
	}

	public final void setDepCompatible(boolean depCompatible) {
		this.depCompatible = depCompatible;
	}

	public final boolean isSafeSehCompatible() {
		return this.safeSehCompatible;
	}

	public final void setSafeSehCompatible(boolean safeSehCompatible) {
		if (!safeSehCompatible){
			if (!getABI().getArchitecture().equals(Architecture.X86)) {
				throw new IllegalArgumentException(String.format("Safe Structural Exception Handling (SEH) can not be disabled on %s architecture", getABI().getArchitecture().toString()));
			}
		}
		this.safeSehCompatible = safeSehCompatible;
	}

	@Override
	public Linker copy() {
		return new MicrosoftLinker(this);
	}

	private final AbsoluteFilePath linkerPath;
	private final List<AbsoluteDirectoryPath> additionalPathDirectories;
	private final List<AbsoluteDirectoryPath> defaultGlobalLibraryDirectories;
	private boolean largeAddressCompatible = true;
	private boolean depCompatible = true;
	private boolean safeSehCompatible = true;
}
