/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.internals.Program;

import java.util.*;
import java.util.regex.Pattern;

public class MicrosoftResourceCompiler {

	public static class OutputParser implements Program.EventHandler {
		OutputParser(BuildMessages buildMessages, AbsoluteFilePath sourcePath, boolean verboseBuild) {
			this.buildMessages = buildMessages;
			this.verboseBuild = verboseBuild;
			this.sourcePath = sourcePath;
			this.filename = sourcePath.getRelativePath(sourcePath.getDirectory()).toString();
		}

		private final BuildMessages buildMessages;
		private final boolean verboseBuild;
		private final AbsoluteFilePath sourcePath;
		private final String filename;

		final Pattern firstLogoLinePattern = Pattern.compile("Microsoft \\(R\\) Windows \\(R\\) Resource Compiler Version (\\d+).(\\d+).(\\d+).(\\d+)");
		final Pattern secondLogoLinePattern = Pattern.compile("Copyright \\(C\\) Microsoft Corporation\\.  All rights reserved\\.");

		@Override
		public void onStdOutLineWrite(String line) {
			if (line.isEmpty()) {
				return;
			} else if (firstLogoLinePattern.matcher(line).matches()) {
				return;
			} else if (secondLogoLinePattern.matcher(line).matches()) {
				return;
			}
			buildMessages.add(new BuildMessage(BuildMessage.Type.Error, sourcePath, null, null, line));
		}

		@Override
		public void onStdErrLineWrite(String line) {
			buildMessages.add(new BuildMessage(BuildMessage.Type.Error, sourcePath, null, null, line));
		}

		@Override
		public void onError(String description) {
			buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, sourcePath, null, null, description));
		}

		@Override
		public void onCompletion(Integer status) {
		}

	}

	public MicrosoftResourceCompiler(AbsoluteFilePath resourceCompilerPath, Collection<AbsoluteDirectoryPath> additionalPathDirectories, Collection<AbsoluteDirectoryPath> defaultIncludeDirectories) {
		this.resourceCompilerPath = resourceCompilerPath;
		this.additionaPathDirectories = new ArrayList<AbsoluteDirectoryPath>(additionalPathDirectories);
		this.defaultIncludeDirectories = new ArrayList<AbsoluteDirectoryPath>(defaultIncludeDirectories);
	}

	public AbsoluteFilePath getObjectPath(AbsoluteFilePath sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (objectDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final RelativeFilePath relativeSourcePath = sourcePath.getRelativePath(sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		return new AbsoluteFilePath(objectDirectory,  relativeSourcePath.replaceExtension("res"));	}

	public BuildMessages compile(final AbsoluteFilePath source) {
		final BuildMessages buildMessages = new BuildMessages();
		final AbsoluteFilePath object = getObjectPath(source);
		if (!object.getDirectory().create()) {
			object.tryDelete();
		}

		final Program.Parameters parameters = prepareParameters(source, object);
		final Program.Environment environment = prepareEnvironment();

		final Program.EventHandler eventHandler = new OutputParser(buildMessages, source, isVerboseBuild());
		if (isVerboseBuild()) {
			System.out.println(resourceCompilerPath + " " + parameters.toString());
		}
		final Program compiler = new Program(resourceCompilerPath, parameters, environment, eventHandler);
		compiler.join();
		return buildMessages;
	}

	private Program.Parameters prepareParameters(AbsoluteFilePath scriptPath, AbsoluteFilePath resourcePath) {
		final Program.Parameters parameters = new Program.Parameters();
		for (final AbsoluteDirectoryPath includeDirectory : this.includeDirectories) {
			parameters.add("/I", includeDirectory.getCanonicalForm());
		}
		parameters.add("/Fo", resourcePath.getCanonicalForm());
		parameters.add(scriptPath.getCanonicalForm());
		return parameters;
	}

	private Program.Environment prepareEnvironment() {
		final Program.Environment environment =  new Program.Environment();
		for (final AbsoluteDirectoryPath additionalPathDirectory : this.additionaPathDirectories) {
			environment.prependPathVariable("PATH", additionalPathDirectory.getCanonicalForm());
		}
		for (final AbsoluteDirectoryPath globalIncludeDirectory : this.globalIncludeDirectories) {
			environment.prependPathVariable("INCLUDE", globalIncludeDirectory.getCanonicalForm());
		}
		return environment;
	}

	public void setObjectDirectory(AbsoluteDirectoryPath objectDirectory) {
		this.objectDirectory = objectDirectory;
	}

	public AbsoluteDirectoryPath getObjectDirectory() {
		return this.objectDirectory;
	}

	public void setSourceDirectory(AbsoluteDirectoryPath sourceDirectory) {
		this.sourceDirectory = sourceDirectory;
	}

	public AbsoluteDirectoryPath getSourceDirectory() {
		return this.sourceDirectory;
	}

	public void setVerboseBuild(boolean verboseBuild) {
		this.verboseBuild = verboseBuild;
	}

	public boolean isVerboseBuild() {
		return this.verboseBuild;
	}

	public boolean addIncludeDirectory(AbsoluteDirectoryPath includePath) {
		return includeDirectories.add(includePath);
	}

	public boolean removeIncludeDirectory(AbsoluteDirectoryPath includePath) {
		return includeDirectories.remove(includePath);
	}

	public void clearIncludeDirectories() {
		includeDirectories.clear();
	}

	public boolean addGlobalIncludeDirectory(AbsoluteDirectoryPath includePath) {
		return globalIncludeDirectories.add(includePath);
	}

	public boolean removeGlobalIncludeDirectory(AbsoluteDirectoryPath includePath) {
		return globalIncludeDirectories.remove(includePath);
	}

	public void clearGlobalIncludeDirectories() {
		globalIncludeDirectories.clear();
	}

	public void addDefaultGlobalIncludeDirectories() {
		for (final AbsoluteDirectoryPath defaultIncludeDirectory : this.defaultIncludeDirectories) {
			this.addGlobalIncludeDirectory(defaultIncludeDirectory);
		}
	}

	public void clearAll() {
		this.clearIncludeDirectories();
		this.clearGlobalIncludeDirectories();
	}

	private AbsoluteDirectoryPath sourceDirectory;
	private AbsoluteDirectoryPath objectDirectory;
	private boolean verboseBuild = false;
	private final Set<AbsoluteDirectoryPath> globalIncludeDirectories = new HashSet<AbsoluteDirectoryPath>();
	private final Set<AbsoluteDirectoryPath> includeDirectories = new HashSet<AbsoluteDirectoryPath>();
	private final AbsoluteFilePath resourceCompilerPath;
	private final List<AbsoluteDirectoryPath> additionaPathDirectories;
	private final List<AbsoluteDirectoryPath> defaultIncludeDirectories;
}
