/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import java.util.regex.Pattern;

public class Version implements Comparable<Version> {
	final static Pattern regex = Pattern.compile("\\d+(?:\\.\\d+)*");

	public Version(String version) {
		if (isValid(version)) {
			int previousComponentIndex = -1;
			int componentsCount = 0;
			do {
				previousComponentIndex = version.indexOf('.', previousComponentIndex + 1);
				componentsCount += 1;
			} while (previousComponentIndex != -1);
			components = new int[componentsCount];
			int start = 0;
			int end = 0;
			int componentIndex = 0;
			do {
				end = version.indexOf('.', start);
				if (end == -1) {
					final String component = version.substring(start);
					components[componentIndex++] = Integer.parseInt(component);
				} else {
					final String component = version.substring(start, end);
					components[componentIndex++] = Integer.parseInt(component);
					start = end + 1;
				}
			} while (end != -1);
			this.version = version.intern();
		} else {
			throw new IllegalArgumentException(String.format("The string %s is not a version string", version));
		}
	}

	public Version(int major) {
		if (major < 0) {
			throw new IllegalArgumentException("The major version must be non-negative");
		}
		this.components = new int[1];
		this.components[0] = major;
		this.version = Integer.toString(major);

	}

	public Version(int major, int minor) {
		if (major < 0) {
			throw new IllegalArgumentException("The major version must be non-negative");
		}
		if (minor < 0) {
			throw new IllegalArgumentException("The minor version must be non-negative");
		}
		this.components = new int[2];
		this.components[0] = major;
		this.components[1] = minor;
		this.version = String.format("%d.%d", major, minor);

	}

	public static boolean isValid(String version) {
		return regex.matcher(version).matches();
	}

	@Override
	public int hashCode() {
		int hashCode = this.components.length;
		for (int i = 0; i < this.components.length; i++) {
			hashCode ^= this.components[i];
		}
		return hashCode;
	}

	@Override
	public String toString() {
		return this.version;
	}

	@Override
	public int compareTo(Version other) {
		if (other.components.length < this.components.length) {
			return -other.compareTo(this);
		} else {
			assert this.components.length < other.components.length;

			for (int componentIndex = 0; componentIndex < this.components.length; componentIndex++) {
				if (this.components[componentIndex] > other.components[componentIndex]) {
					return 1;
				} else if (this.components[componentIndex] < other.components[componentIndex]) {
					return -1;
				}
			}
			for (int componentIndex = this.components.length; componentIndex < other.components.length; componentIndex++) {
				if (other.components[componentIndex] > 0) {
					return -1;
				}
			}
			return 0;
		}
	}

	private final String version;
	private final int[] components;
}
