/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda;

import info.yeppp.ebuilda.filesystem.AbsoluteDirectoryPath;
import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;
import info.yeppp.ebuilda.filesystem.RelativeDirectoryPath;
import info.yeppp.ebuilda.filesystem.RelativeFilePath;
import info.yeppp.ebuilda.generic.Versioned;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

public abstract class VisualStudio implements Versioned {

	public static List<VisualStudio> enumerate(Machine machine) {
		final List<VisualStudio> implementations = new LinkedList<VisualStudio>();
		implementations.addAll(VisualStudio2012.enumerate(machine));
		implementations.addAll(VisualStudio2010.enumerate(machine));
		implementations.addAll(VisualStudio2008.enumerate(machine));
		return implementations;
	}

	public static VersionList<VisualStudio> enumerate(Machine machine, ABI abi) {
		final VersionList<VisualStudio> implementations = new VersionList<VisualStudio>();
		implementations.add(VisualStudio2012.enumerate(machine, abi));
		implementations.add(VisualStudio2010.enumerate(machine, abi));
		implementations.add(VisualStudio2008.enumerate(machine, abi));
		return implementations;
	}

	protected VisualStudio(AbsoluteDirectoryPath rootDirectory, AbsoluteDirectoryPath cppIncludeDirectory, AbsoluteDirectoryPath cppLibraryDirectory, AbsoluteDirectoryPath cppToolsDirectory, WindowsSDK windowsSDK, ABI abi, Version version) {
		this.rootDirectory = rootDirectory;
		this.cppIncludeDirectory = cppIncludeDirectory;
		this.cppLibraryDirectory = cppLibraryDirectory;
		this.cppToolsDirectory = cppToolsDirectory;
		this.windowsSDK = windowsSDK;
		this.abi = abi;
		this.version = version;

		final List<AbsoluteDirectoryPath> additionalPathDirectories = new ArrayList<AbsoluteDirectoryPath>(3);
		additionalPathDirectories.add(getCppToolsDirectory());
		additionalPathDirectories.add(getCommonToolsDirectory());
		additionalPathDirectories.add(getIdeDirectory());
		final Collection<AbsoluteDirectoryPath> defaultIncludeDirectories = getCppIncludeDirectories();
		defaultIncludeDirectories.addAll(windowsSDK.getIncludeDirectories());
		final Collection<AbsoluteDirectoryPath> defaultLibraryDirectories = getCppLibraryDirectories();
		defaultLibraryDirectories.addAll(windowsSDK.getLibraryDirectories());

		final AbsoluteFilePath cppCompilerPath = new AbsoluteFilePath(getCppToolsDirectory(), new RelativeFilePath("cl.exe"));
		final Version cppCompilerVersion = MicrosoftCppCompiler.verify(cppCompilerPath, additionalPathDirectories).getVersion();

		this.cppCompiler = new MicrosoftCppCompiler(cppCompilerPath, abi, cppCompilerVersion, additionalPathDirectories, defaultIncludeDirectories);
		this.cCompiler = new MicrosoftCCompiler(cppCompilerPath, abi, cppCompilerVersion, additionalPathDirectories, defaultIncludeDirectories);
		this.linker = new MicrosoftLinker(new AbsoluteFilePath(getCppToolsDirectory(), new RelativeFilePath("link.exe")),
			abi, additionalPathDirectories, defaultLibraryDirectories);
	}

	protected static final WindowsSDK findWindowsSDK(Machine machine, ABI abi) {
		try {
			return WindowsSDK.enumerate(machine, abi).get(0);
		} catch (IndexOutOfBoundsException e) {
			throw new Error(String.format("No Windows SDK found to use with Visual Studio on %s", abi.toString()));
		}
	}

	public MicrosoftCppCompiler getCppCompiler() {
		return this.cppCompiler;
	}

	public MicrosoftCCompiler getCCompiler() {
		return this.cCompiler;
	}

	public MicrosoftLinker getLinker() {
		return this.linker;
	}

	public WindowsSDK getWindowsSDK() {
		return this.windowsSDK;
	}

	public final AbsoluteDirectoryPath getRootDirectory() {
		return this.rootDirectory;
	}

	public Collection<AbsoluteDirectoryPath> getCppIncludeDirectories() {
		final List<AbsoluteDirectoryPath> cppIncludeDirectories = new LinkedList<AbsoluteDirectoryPath>();
		cppIncludeDirectories.add(this.cppIncludeDirectory);
		return cppIncludeDirectories;
	}

	public Collection<AbsoluteDirectoryPath> getCppLibraryDirectories() {
		final List<AbsoluteDirectoryPath> cppLibraryDirectories = new LinkedList<AbsoluteDirectoryPath>();
		cppLibraryDirectories.add(this.cppLibraryDirectory);
		return cppLibraryDirectories;
	}

	public AbsoluteDirectoryPath getCppToolsDirectory() {
		return this.cppToolsDirectory;
	}

	public AbsoluteDirectoryPath getIdeDirectory() {
		return new AbsoluteDirectoryPath(rootDirectory, new RelativeDirectoryPath("Common7\\IDE"));
	}

	public AbsoluteDirectoryPath getCommonToolsDirectory() {
		return new AbsoluteDirectoryPath(rootDirectory, new RelativeDirectoryPath("Common7\\Tools"));
	}

	public final ABI getABI() {
		return this.abi;
	}

	@Override
	public final Version getVersion() {
		return this.version;
	}

	@Override
	public String toString() {
		return String.format("Visual Studio version %s for %s", this.version.toString(), this.abi.toString());
	}

	private final ABI abi;
	private final Version version;
	private WindowsSDK windowsSDK;
	private final AbsoluteDirectoryPath rootDirectory;
	private final AbsoluteDirectoryPath cppIncludeDirectory;
	private final AbsoluteDirectoryPath cppLibraryDirectory;
	private final AbsoluteDirectoryPath cppToolsDirectory;
	private final MicrosoftCppCompiler cppCompiler;
	private final MicrosoftCCompiler cCompiler;
	private final MicrosoftLinker linker;
}
