/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda.filesystem;

import info.yeppp.ebuilda.Machine;
import info.yeppp.ebuilda.OperatingSystem;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.regex.Pattern;

public final class AbsoluteDirectoryPath {

	public AbsoluteDirectoryPath(Machine machine, String path) {
		this.machine = machine;
		if (path == null) {
			throw new NullPointerException("The path string is null");
		}
		if (path.isEmpty()) {
			throw new PathFormatException("The path string is empty");
		}
		/* The internal representation always uses POSIX file separators.
		 * Convert Windows path separators '/' into POSIX path separators '\' */
		path = path.replace('\\', '/');
		if (getMachine().getOperatingSystem().equals(OperatingSystem.Windows)) {
			if (Pattern.matches("[A-Za-z]:/.*", path)) {
				/* Uppercase the drive letter */
				path = path.substring(0, 1).toUpperCase(Locale.US) + path.substring(1);
				final String driveRelativePath = path.substring(3);
				/* Check for forbidden symbols on Windows
				 * Information about forbidden symbols from
				 * http://msdn.microsoft.com/en-us/library/windows/desktop/aa365247.aspx
				 */
				checkForForbiddenWindowsPathSymbol(driveRelativePath, '<');
				checkForForbiddenWindowsPathSymbol(driveRelativePath, '>');
				checkForForbiddenWindowsPathSymbol(driveRelativePath, ':');
				checkForForbiddenWindowsPathSymbol(driveRelativePath, '"');
				checkForForbiddenWindowsPathSymbol(driveRelativePath, '|');
				checkForForbiddenWindowsPathSymbol(driveRelativePath, '?');
				checkForForbiddenWindowsPathSymbol(driveRelativePath, '*');
				final String[] components = path.split("/");

				/* Remove empty directories (C:///directory/),
				 * references to current directory (C:/./directory/),
				 * and references to the parent directory (C:/parent/..\directory/)
				 */
				final List<String> directories = new ArrayList<String>(components.length);
				for (int i = 1; i < components.length; i++) {
					final String component = components[i];
					if (component.equals(".")) {
						continue;
					} else if (component.equals("..")) {
						if (directories.isEmpty()) {
							/* Reference to parent directory, but no actual parent directory in the path (C:/parent/../../directory/) */
							throw new PathFormatException(String.format("The path %s has imbalanced references to parent directories (..)", path));
						} else {
							directories.remove(directories.size() - 1);
						}
					} else {
						directories.add(component);
					}
				}
				/* Make the final path */
				final StringBuilder stringBuilder = new StringBuilder(directories.size() * 2 + 1);
				stringBuilder.append(path.substring(0, 3));
				for (final String directory : directories) {
					stringBuilder.append(directory);
					stringBuilder.append('/');
				}
				this.path = stringBuilder.toString();
			} else {
				throw new PathFormatException(String.format("The string %s is not an absolute directory path on Windows: disk drive is not specified", path));
			}
		} else {
			if (path.startsWith("//")) {
				throw new PathFormatException(String.format("The UNC path %s is not supported", path));
			}

			if (path.startsWith("/")) {
				checkForForbiddenUnixPathSymbol(path, '<');
				checkForForbiddenUnixPathSymbol(path, '>');
				checkForForbiddenUnixPathSymbol(path, '"');
				checkForForbiddenUnixPathSymbol(path, '|');
				checkForForbiddenUnixPathSymbol(path, '?');
				checkForForbiddenUnixPathSymbol(path, '*');
				final String[] components = path.split("/");

				/* Remove empty directories (/home/user//filename.txt),
				 * references to current directory (/home/user/./filename.txt),
				 * and references to the parent directory (/home/user/parent/../filename.txt)
				 */
				final List<String> directories = new ArrayList<String>(components.length);
				for (int i = 1; i < components.length; i++) {
					final String component = components[i];
					if (component.equals(".") || component.isEmpty()) {
						continue;
					} else if (component.equals("..")) {
						if (directories.isEmpty()) {
							/* Reference to parent directory, but no actual parent directory in the path (/home/user/parent/../../../../filename.txt) */
							throw new PathFormatException(String.format("The path %s has imbalanced references to parent directories (..)", path));
						} else {
							directories.remove(directories.size() - 1);
						}
					} else {
						directories.add(component);
					}
				}
				/* Make the final path */
				final StringBuilder stringBuilder = new StringBuilder(directories.size() * 2 + 1);
				stringBuilder.append('/');
				for (final String directory : directories) {
					stringBuilder.append(directory);
					stringBuilder.append('/');
				}
				this.path = stringBuilder.toString();
			} else {
				throw new PathFormatException(String.format("The string %s is not an absolute file path on Unix", path));
			}
		}
	}

	public AbsoluteDirectoryPath(AbsoluteDirectoryPath directory, RelativeDirectoryPath subdirectory) {
		String directoryPath = directory.toString();
		String subdirectoryPath = subdirectory.toString();
		while (subdirectoryPath.startsWith("../")) {
			final int separatorIndex = directoryPath.lastIndexOf('/', directoryPath.length() - 2); /* Skip the final separator */
			if (separatorIndex == -1) {
				throw new PathFormatException(String.format("The relative directory path %s has too many references to parent directories to combine with directory path %s", subdirectory.toString(), directory.toString()));
			} else {
				directoryPath = directoryPath.substring(0, separatorIndex + 1);
				subdirectoryPath = subdirectoryPath.substring(3);
			}
		}
		this.machine = directory.machine;
		this.path = directoryPath + subdirectoryPath;

	}

	public AbsoluteDirectoryPath(AbsoluteDirectoryPath other) {
		this.machine = other.machine;
		this.path = other.path;
	}

	public AbsoluteDirectoryPath resolve() {
		try {
			return new AbsoluteDirectoryPath(this.getMachine(), new File(this.getCanonicalForm()).getCanonicalPath());
		} catch (IOException e) {
			return this;
		}
	}

	public static boolean isValid(final Machine machine, final String path) {
		if (path == null) {
			return false;
		} else {
			try {
				new AbsoluteDirectoryPath(machine, path);
				return true;
			} catch (PathFormatException e) {
				return false;
			}
		}
	}

	public boolean create() {
		if (getMachine() == Machine.getLocal()) {
			final File directory = new File(getCanonicalForm());
			if (directory.mkdirs()) {
				return true;
			} else if (directory.exists() && directory.isDirectory() ) {
				return false;
			} else {
				throw new Error(String.format("Could not create the directory %s", toString()));
			}
		} else {
			throw new Error("Remote connection is not supported yet");
		}
	}

	public AbsoluteDirectoryPath getParent() {
		return new AbsoluteDirectoryPath(this, new RelativeDirectoryPath(".."));
	}

	public boolean exists() {
		if (getMachine() == Machine.getLocal()) {
			final File directory = new File(getCanonicalForm());
			return directory.exists() && directory.isDirectory();
		} else {
			throw new Error("Remote connection is not supported yet");
		}
	}

	public boolean contains(RelativeFilePath relativePath) {
		if (getMachine() == Machine.getLocal()) {
			final AbsoluteFilePath absolutePath = new AbsoluteFilePath(this, relativePath);
			return absolutePath.exists();
		} else {
			throw new Error("Remote connection is not supported yet");
		}
	}

	public boolean contains(RelativeDirectoryPath relativePath) {
		if (getMachine() == Machine.getLocal()) {
			final AbsoluteDirectoryPath absolutePath = new AbsoluteDirectoryPath(this, relativePath);
			return absolutePath.exists();
		} else {
			throw new Error("Remote connection is not supported yet");
		}
	}

	private static void listFiles(AbsoluteDirectoryPath directory, List<AbsoluteFilePath> filenames, final Pattern pattern, final int maxRecursion, final boolean executablesOnly, final boolean matchBaseName) {
		if (directory.exists()) {
			final File directoryFile = new File(directory.getCanonicalForm());
			final File[] files = directoryFile.listFiles(new FileFilter() {
				@Override
				public boolean accept(File file) {
					if (maxRecursion > 0) {
						if (file.isHidden()) {
							return false;
						} else {
							if (file.isDirectory()) {
								return true;
							} else {
								if (executablesOnly && !file.canExecute()) {
									return false;
								}
								String name = file.getName();
								if (matchBaseName) {
									final int extensionSeparator = name.lastIndexOf('.');
									if (extensionSeparator != -1) {
										name = name.substring(0, extensionSeparator);
									}
								}
								return pattern.matcher(name).matches();
							}
						}
					} else {
						String name = file.getName();
						if (matchBaseName) {
							final int extensionSeparator = name.lastIndexOf('.');
							if (extensionSeparator != -1) {
								name = name.substring(0, extensionSeparator);
							}
						}
						if (pattern.matcher(name).matches()) {
							return file.canExecute() && file.isFile() && !file.isHidden();
						} else {
							return false;
						}
					}
				}
			});
			for (int i = 0; i < files.length; i++) {
				final File file = files[i];
				if ((maxRecursion == 0) || file.isFile()) {
					filenames.add(new AbsoluteFilePath(directory.getMachine(), file.getAbsolutePath()));
				}
			}
			if (maxRecursion > 0) {
				for (int i = 0; i < files.length; i++) {
					final File file = files[i];
					if (file.isDirectory()) {
						final AbsoluteDirectoryPath subdirectory = new AbsoluteDirectoryPath(directory.getMachine(), file.getAbsolutePath());
						listFiles(subdirectory, filenames, pattern, maxRecursion - 1, executablesOnly, matchBaseName);
					}
				}
			}
		}
	}

	public List<AbsoluteFilePath> getFiles(Pattern pattern, boolean recursively) {
		final List<AbsoluteFilePath> files = new ArrayList<AbsoluteFilePath>();
		listFiles(this, files, pattern, recursively ? 25 : 0, false, false);
		return files;
	}

	public List<AbsoluteFilePath> getExecutables(Pattern pattern, boolean recursively) {
		final List<AbsoluteFilePath> files = new ArrayList<AbsoluteFilePath>();
		listFiles(this, files, pattern, recursively ? 25 : 0, true, true);
		return files;
	}

	public List<AbsoluteDirectoryPath> getDirectories(final Pattern pattern) {
		final File directoryFile = new File(getCanonicalForm());
		final File[] files = directoryFile.listFiles(new FileFilter() {
			@Override
			public boolean accept(File file) {
				return file.isDirectory() && !file.isHidden() && pattern.matcher(file.getName()).matches();
			}
		});
		final List<AbsoluteDirectoryPath> filenames = new ArrayList<AbsoluteDirectoryPath>(files.length);
		for (final File file : files) {
			filenames.add(new AbsoluteDirectoryPath(getMachine(), file.getAbsolutePath()));
		}
		return filenames;
	}

	public String getLastFileName() {
		final int lastSeparator = this.path.lastIndexOf('/', this.path.length() - 2);
		if (lastSeparator == -1) {
			throw new Error(String.format("The path %s does not contain any path separator '/'", this.path));
		} else {
			return this.path.substring(lastSeparator + 1, this.path.length() - 1);
		}
	}

	private void checkForForbiddenWindowsPathSymbol(final String path, final char symbol) {
		final String symbolString = new String(new char[]{symbol});
		if (path.contains(symbolString)) {
			throw new PathFormatException(String.format("The directory path %s contains symbol '%c' which is not allowed on Windows", path, symbol));
		}

	}

	private void checkForForbiddenUnixPathSymbol(final String path, final char symbol) {
		final String symbolString = new String(new char[]{symbol});
		if (path.contains(symbolString)) {
			throw new PathFormatException(String.format("The directory path %s contains symbol '%c' which is not allowed on Unix", path, symbol));
		}

	}

	protected final Machine machine;
	private final String path;

	public String getCanonicalForm() {
		if (getMachine().getOperatingSystem().equals(OperatingSystem.Windows)) {
			return this.path.replace('/', '\\');
		} else {
			return this.path;
		}
	}

	public Machine getMachine() {
		return this.machine;
	}

	@Override
	public String toString() {
		return this.path;
	}

	@Override
	public boolean equals(Object other) {
		if (other == null) {
			return false;
		} else if (other instanceof AbsoluteDirectoryPath) {
			final AbsoluteDirectoryPath otherPath = (AbsoluteDirectoryPath)other;
			return this.path.equals(otherPath.path) && this.machine.equals(otherPath.machine);
		} else if (other instanceof String) {
			return this.path.equals(other) && this.machine.isLocal();
		} else {
			return false;
		}
	}

	@Override
	public int hashCode() {
		return this.path.hashCode() ^ this.machine.hashCode();
	}
}
