/*
 *                      EBUILDA: cross-platform build system
 *
 * This file is part of Yeppp! library infrastructure and licensed under Simplified BSD License.
 * See LICENSE.txt for the full text of the license.
 *
 */

package info.yeppp.ebuilda.internals;

import info.yeppp.ebuilda.filesystem.AbsoluteFilePath;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.*;
import java.util.regex.Pattern;

public class Program {

	public static interface EventHandler {
		public void onStdOutLineWrite(String line);
		public void onStdErrLineWrite(String line);
		public void onError(String description);
		public void onCompletion(Integer status);
	}

	public static class Environment {

		public Environment() {
			variables = new HashMap<String, String>(System.getenv());
		}

		public boolean addVariable(String name, String value) {
			if (variables.containsKey(name)) {
				return false;
			} else {
				variables.put(name, value);
				return true;
			}
		}

		public boolean appendPathVariable(String name, String path) {
			final String variable = getVariable(name);
			if (variable.isEmpty()) {
				setVariable(name, path);
				return true;
			} else {
				final String paths[] = variable.split(Pattern.quote(pathSeparator));
				for (int i = 0; i < paths.length; i++) {
					if (paths[i].equals(path)) {
						return false;
					}
				}
				setVariable(name, variable + pathSeparator + path);
				return true;
			}
		}

		public boolean prependPathVariable(String name, String path) {
			final String variable = getVariable(name);
			if (variable.isEmpty()) {
				variables.put(name, path);
				return true;
			} else {
				final String paths[] = variable.split(Pattern.quote(pathSeparator));
				for (int i = 0; i < paths.length; i++) {
					if (paths[i].equals(path)) {
						return false;
					}
				}
				setVariable(name, path + pathSeparator + variable);
				return true;
			}
		}

		public boolean setVariable(String name, String value) {
			for (final String variable : variables.keySet()) {
				if (variable.equalsIgnoreCase(name)) {
					variables.put(variable, value);
					return true;
				}
			}
			variables.put(name, value);
			return false;
		}

		public String getVariable(String name) {
			for (final String variable : variables.keySet()) {
				if (variable.equalsIgnoreCase(name)) {
					return variables.get(variable);
				}
			}
			return new String();
		}

		public void clear() {
			variables.clear();
			variables.putAll(System.getenv());
		}

		final String pathSeparator = System.getProperty("path.separator");
		final private Map<String, String> variables;
	}

	private static class Message {
		private static enum Type {
			StdOutLineWrite,
			StdErrLineWrite,
			Error,
			Completion
		}

		public Message(Type type, Object content) {
			this.type = type;
			this.content = content;
		}

		final Type type;
		final Object content;
	}

	public static class Parameters {
		public Parameters() {
			this.parameters = new LinkedList<String>();
		}

		public Parameters(String parameter) {
			this.parameters = new LinkedList<String>();
			this.parameters.add(parameter);
		}

		public Parameters(String parameter1, String parameter2) {
			this.parameters = new LinkedList<String>();
			this.parameters.add(parameter1);
			this.parameters.add(parameter2);
		}

		public Parameters(String parameter1, String parameter2, String parameter3) {
			this.parameters = new LinkedList<String>();
			this.parameters.add(parameter1);
			this.parameters.add(parameter2);
			this.parameters.add(parameter3);
		}

		public void add(String parameter) {
			this.parameters.add(parameter);
		}

		public void add(String parameter1, String parameter2) {
			this.parameters.add(parameter1);
			this.parameters.add(parameter2);
		}

		public void add(String parameter1, String parameter2, String parameter3) {
			this.parameters.add(parameter1);
			this.parameters.add(parameter2);
			this.parameters.add(parameter3);
		}

		public void add(String parameter1, String parameter2, String parameter3, String parameter4) {
			this.parameters.add(parameter1);
			this.parameters.add(parameter2);
			this.parameters.add(parameter3);
			this.parameters.add(parameter4);
		}

		public int count() {
			return parameters.size();
		}

		@Override
		public String toString() {
			final StringBuilder output = new StringBuilder(this.parameters.size() * 2);
			final Iterator<String> iterator = parameters.iterator();
			String parameter = iterator.next();
			if (parameter != null) {
				output.append(parameter);
			}
			while (iterator.hasNext()) {
				parameter = iterator.next();
				output.append(' ');
				output.append(parameter);
			}
			return output.toString();
		}

		List<String> parameters;
	}

	private static final class MessageQueue {
		public MessageQueue() {
			this.messages = new LinkedList<Message>();
		}

		public synchronized void addMessage(Message.Type type, String text) {
			assert type != Message.Type.Completion;
			assert !this.processedCompletionMessage;
			this.messages.add(new Message(type, text));
		}

		public synchronized void addCompletionMessage(Integer statusCode) {
			assert !this.processedCompletionMessage;
			if (statusCode != null) {
				this.statusCode = statusCode;
			}
			this.hasCompletionMessage = true;
		}

		public synchronized boolean hasMessage() {
			if (this.processedCompletionMessage) {
				return false;
			} else if (this.hasCompletionMessage) {
				return true;
			} else {
				return !this.messages.isEmpty();
			}
		}

		public synchronized Message getMessage() {
			assert !this.processedCompletionMessage;
			synchronized (this) {
				if (this.messages.isEmpty()) {
					if (this.hasCompletionMessage) {
						this.processedCompletionMessage = true;
						this.hasCompletionMessage = false;
						return new Message(Message.Type.Completion, this.statusCode);
					} else {
						return null;
					}
				} else {
					return this.messages.remove(0);
				}
			}
		}

		public synchronized Message peekMessage() {
			assert !this.processedCompletionMessage;
			if (this.messages.isEmpty()) {
				if (this.hasCompletionMessage) {
					return new Message(Message.Type.Completion, this.statusCode);
				} else {
					return null;
				}
			} else {
				return this.messages.get(0);
			}
		}

		private boolean hasCompletionMessage = false;
		private boolean processedCompletionMessage = false;
		private Integer statusCode = null;
		private final List<Message> messages;
	}

	private static final class ProcessWaitThread extends Thread {
		public ProcessWaitThread(java.lang.Process process, MessageQueue messageQueue) {
			this.process = process;
			this.messageQueue = messageQueue;
		}

		@Override
		public void run() {
			try {
				final int status = process.waitFor();
				messageQueue.addCompletionMessage(status);
				synchronized (messageQueue) {
					messageQueue.notifyAll();
				}
			} catch (InterruptedException exception) {
				process.destroy();
				messageQueue.addMessage(Message.Type.Error, exception.getMessage());
				messageQueue.addCompletionMessage(null);
				synchronized (messageQueue) {
					messageQueue.notifyAll();
				}
			}
		}

		private final java.lang.Process process;
		private final MessageQueue messageQueue;
	}

	private static final class ProcessReadThread extends Thread {
		public ProcessReadThread(InputStream stream, boolean isErrorStream, MessageQueue messageQueue) {
			this.reader = new BufferedReader(new InputStreamReader(stream, Charset.defaultCharset()));
			this.isErrorStream = isErrorStream;
			this.messageQueue = messageQueue;
		}

		@Override
		public void run() {
			String line = null;
			try {
				while ((line = this.reader.readLine()) != null) {
					if (this.isErrorStream) {
						messageQueue.addMessage(Message.Type.StdErrLineWrite, line);
					} else {
						messageQueue.addMessage(Message.Type.StdOutLineWrite, line);
					}
					synchronized (messageQueue) {
						messageQueue.notifyAll();
					}
				}
			} catch (Exception e) {
				System.err.println(e.getMessage());
			}
		}

		final BufferedReader reader;
		final boolean isErrorStream;
		final MessageQueue messageQueue;
	}

	public static List<String> getStdoutLines(AbsoluteFilePath programPath, Parameters parameters, Environment environment, final int linesCount) {
		final List<String> lines = new ArrayList<String>(linesCount);
		final Program process = new Program(programPath, parameters, environment, new EventHandler() {
			@Override
			public void onStdOutLineWrite(String line) {
				if (lines.size() < linesCount) {
					lines.add(line);
				}
			}

			@Override
			public void onStdErrLineWrite(String line) {
			}

			@Override
			public void onError(String description) {
			}

			@Override
			public void onCompletion(Integer status) {
			}
		});
		process.join();
		return lines;
	}

	public static List<String> getStderrLines(AbsoluteFilePath programPath, Parameters parameters, Environment environment, final int linesCount) {
		final List<String> lines = new ArrayList<String>(linesCount);
		final Program process = new Program(programPath, parameters, environment, new EventHandler() {
			@Override
			public void onStdOutLineWrite(String line) {
			}

			@Override
			public void onStdErrLineWrite(String line) {
				if (lines.size() < linesCount) {
					lines.add(line);
				}
			}

			@Override
			public void onError(String description) {
			}

			@Override
			public void onCompletion(Integer status) {
			}
		});
		process.join();
		return lines;
	}

	public Program(AbsoluteFilePath programPath, Parameters parameters, Environment environment, EventHandler eventHandler) {
		String[] commands = new String[parameters.count() + 1];
		String[] variables = new String[environment.variables.size()];
		{
			int index = 0;
			commands[index++] = programPath.getCanonicalForm();
			for (final String parameter : parameters.parameters) {
				commands[index++] = parameter;
			}
		}
		{
			int index = 0;
			for (final Map.Entry<String, String> variable : environment.variables.entrySet()) {
				variables[index++] = variable.getKey() + "=" + variable.getValue();
			}
		}
		this.messageQueue = new MessageQueue();
		this.eventHandler = eventHandler;
		try {
			this.process = Runtime.getRuntime().exec(commands, variables);
			this.stdErrReadThread = new ProcessReadThread(process.getInputStream(), false, this.messageQueue);
			this.stdOutReadThread = new ProcessReadThread(process.getErrorStream(), true, this.messageQueue);
			this.processWaitThread = new ProcessWaitThread(process, this.messageQueue);

			this.stdOutReadThread.start();
			this.stdErrReadThread.start();
			this.processWaitThread.start();
		} catch (IOException exception) {
			this.eventHandler.onError(exception.getMessage());
			this.eventHandler.onCompletion(null);

			this.process = null;
			this.stdErrReadThread = null;
			this.stdOutReadThread = null;
			this.processWaitThread = null;
		}
	}

	public void processMessages() {
		dispatchEvents();
	}

	private void joinThreads() {
		while (stdErrReadThread.isAlive()) {
			try {
				stdErrReadThread.join();
			} catch (InterruptedException e) {
			}
		}
		while (stdOutReadThread.isAlive()) {
			try {
			stdOutReadThread.join();
			} catch (InterruptedException e) {
			}
		}
		while (processWaitThread.isAlive()) {
			try {
				processWaitThread.join();
			} catch (InterruptedException e) {
			}
		}
	}

	public void destroy() {
		process.destroy();
		joinThreads();
	}

	public Integer join() {
		if (this.process == null) {
			return null;
		} else {
				while (!isFinished) {
					try {
						synchronized (messageQueue) {
							messageQueue.wait();
						}
					} catch (InterruptedException exception) {
					}
					dispatchEvents();
				}
			return this.status;
		}
	}

	private void dispatchEvents() {
		if (isFinished) {
			return;
		}

		boolean receivedCompletionMessage = false;
		while (messageQueue.hasMessage()) {
			final Message message = messageQueue.peekMessage();
			switch (message.type) {
				case StdOutLineWrite:
				{
					final Message lastMessage = messageQueue.getMessage();
					assert lastMessage == message;

					final String line = (String) message.content;
					eventHandler.onStdOutLineWrite(line);
					break;
				}
				case StdErrLineWrite:
				{
					final Message lastMessage = messageQueue.getMessage();
					assert lastMessage == message;

					final String line = (String) message.content;
					eventHandler.onStdErrLineWrite(line);
					break;
				}
				case Error:
				{
					final Message lastMessage = messageQueue.getMessage();
					assert lastMessage == message;

					final String description = (String) message.content;
					eventHandler.onError(description);
					break;
				}
				case Completion:
				{
					if (receivedCompletionMessage) {
						final Message lastMessage = messageQueue.getMessage();
						assert !messageQueue.hasMessage();

						final Integer status = message.content == null ? null : (Integer) message.content;
						this.status = status;
						this.isFinished = true;
						eventHandler.onCompletion(status);
					} else {
						joinThreads();
						receivedCompletionMessage = true;
					}
				}
			}
		}
	}

	private boolean isFinished = false;
	private Integer status = null;
	private final EventHandler eventHandler;
	private final MessageQueue messageQueue;
	private java.lang.Process process;
	private Thread stdErrReadThread;
	private Thread stdOutReadThread;
	private Thread processWaitThread;
}
